import re
from typing import Optional, Union, List, Any
from dataclasses import dataclass, field
from datetime import date, datetime
from functools import singledispatch
from dataclasses import dataclass
import numpy as np
import pandas as pd
import astropy.time
import astropy.units as u
from astropy.time import Time

TIME_REGEX = {
    '%Y': r'(?P<year>\d{4})',
    '%j': r'(?P<dayofyear>\d{3})',
    '%m': r'(?P<month>\d{1,2})',
    '%d': r'(?P<day>\d{1,2})',
    '%H': r'(?P<hour>\d{1,2})',
    '%M': r'(?P<minute>\d{1,2})',
    '%S': r'(?P<second>\d{1,2})',
    '%f': r'(?P<microsecond>\d+)',
    '%b': r'(?P<month_str>[a-zA-Z]+)',
}


@dataclass(frozen=True)
class TimeFormats:
    common_formats: List[str] = field(default_factory=lambda: (
        "%Y-%m-%dT%H:%M:%S",  # 2007-05-04T21:08:12
        "%Y/%m/%dT%H:%M:%S",  # 2007/05/04T21:08:12
        "%Y%m%dT%H%M%S",  # 20070504T210812
        "%Y/%m/%d %H:%M:%S",  # 2007/05/04 21:08:12
        "%Y/%m/%d %H:%M",  # 2007/05/04 21:08
        "%Y-%m-%d %H:%M:%S",  # 2007-05-04 21:08:12
        "%Y-%m-%d %H:%M",  # 2007-05-04 21:08
        "%Y-%b-%d %H:%M:%S",  # 2007-May-04 21:08:12
        "%Y-%b-%d %H:%M",  # 2007-May-04 21:08
        "%Y-%b-%d",  # 2007-May-04
        "%Y-%m-%d",  # 2007-05-04
        "%Y/%m/%d",  # 2007/05/04
        "%d-%b-%Y",  # 04-May-2007
        "%d-%b-%Y %H:%M:%S",  # 04-May-2007 21:08:12
        "%Y%m%d_%H%M%S",  # 20070504_210812
        "%Y:%j:%H:%M:%S",  # 2012:124:21:08:12
        "%Y%m%d%H%M%S",  # 20140101000001 (JSOC/VSO Export/Downloads)
        "%Y.%m.%d_%H:%M:%S_TAI",  # 2016.05.04_21:08:12_TAI - JSOC
        "%Y.%m.%d_%H:%M:%S_UTC",  # 2016.05.04_21:08:12_UTC - JSOC
        "%Y.%m.%d_%H:%M:%S",  # 2016.05.04_21:08:12 - JSOC
        "%Y/%m/%dT%H:%M",  # 2007/05/04T21:08
        "%Y-%m-%dT%H:%M:%S.%f",  # 2007-05-04T21:08:12.999999
        "%Y/%m/%dT%H:%M:%S.%f",  # 2007/05/04T21:08:12.999999
        "%Y-%m-%dT%H:%M:%S.%fZ",  # 2007-05-04T21:08:12.999Z
        "%Y%m%dT%H%M%S.%f",  # 20070504T210812.999999
         "%Y/%m/%d %H:%M:%S.%f",  # 2007/05/04 21:08:12.999999
        "%Y-%m-%d %H:%M:%S.%f",  # 2007-05-04 21:08:12.999999
        "%Y-%b-%d %H:%M:%S.%f",  # 2007-May-04 21:08:12.999999
        "%d-%b-%Y %H:%M:%S.%f",  # 04-May-2007 21:08:12.999999
        "%Y:%j:%H:%M:%S.%f",  # 2012:124:21:08:12.999999
    ))
    examples: List[str] = field(default_factory=lambda: (
        "2007-05-04T21:08:12",
        "2007/05/04T21:08:12",
        "20070504T210812",
        "2007/05/04 21:08:12",
        "2007/05/04 21:08",
        "2007-05-04 21:08:12",
        "2007-05-04 21:08",
        "2007-May-04 21:08:12",
        "2007-May-04 21:08",
        "2007-May-04",
        "2007-05-04",
        "2007/05/04",
        "04-May-2007",
        "04-May-2007 21:08:12",
        "20070504_210812",
        "2012:124:21:08:12",
        "20140101000001", #(JSOC/VSO Export/Downloads)
        "2016.05.04_21:08:12_TAI",  #JSOC
        "2016.05.04_21:08:12_UTC",  #JSOC
        "2016.05.04_21:08:12",  #JSOC
        "2007/05/04T21:08",
        "2007-05-04T21:08:12.999999",
        "2007/05/04T21:08:12.999999",
        "2007-05-04T21:08:12.999Z",
        "20070504T210812.999999",
        "2007/05/04 21:08:12.999999",
        "2007-05-04 21:08:12.999999",
        "2007-May-04 21:08:12.999999",
        "04-May-2007 21:08:12.999999",
        "2012:124:21:08:12.999999",
    ))


COMMON_TIME_FORMATS = [
    "%Y-%m-%dT%H:%M:%S",  # 2007-05-04T21:08:12
    "%Y/%m/%dT%H:%M:%S",  # 2007/05/04T21:08:12
    "%Y%m%dT%H%M%S",  # 20070504T210812
    "%Y/%m/%d %H:%M:%S",  # 2007/05/04 21:08:12
    "%Y/%m/%d %H:%M",  # 2007/05/04 21:08
    "%Y-%m-%d %H:%M:%S",  # 2007-05-04 21:08:12
    "%Y-%m-%d %H:%M",  # 2007-05-04 21:08
    "%Y-%b-%d %H:%M:%S",  # 2007-May-04 21:08:12
    "%Y-%b-%d %H:%M",  # 2007-May-04 21:08
    "%Y-%b-%d",  # 2007-May-04
    "%Y-%m-%d",  # 2007-05-04
    "%Y/%m/%d",  # 2007/05/04
    "%d-%b-%Y",  # 04-May-2007
    "%d-%b-%Y %H:%M:%S",  # 04-May-2007 21:08:12
    "%Y%m%d_%H%M%S",  # 20070504_210812
    "%Y:%j:%H:%M:%S",  # 2012:124:21:08:12
    "%Y%m%d%H%M%S",  # 20140101000001 (JSOC/VSO Export/Downloads)
    "%Y.%m.%d_%H:%M:%S_TAI",  # 2016.05.04_21:08:12_TAI - JSOC
    "%Y.%m.%d_%H:%M:%S_UTC",  # 2016.05.04_21:08:12_UTC - JSOC
    "%Y.%m.%d_%H:%M:%S",  # 2016.05.04_21:08:12 - JSOC
    "%Y/%m/%dT%H:%M",  # 2007/05/04T21:08
    "%Y-%m-%dT%H:%M:%S.%f",  # 2007-05-04T21:08:12.999999
    "%Y/%m/%dT%H:%M:%S.%f",  # 2007/05/04T21:08:12.999999
    "%Y-%m-%dT%H:%M:%S.%fZ",  # 2007-05-04T21:08:12.999Z
    "%Y%m%dT%H%M%S.%f",  # 20070504T210812.999999
     "%Y/%m/%d %H:%M:%S.%f",  # 2007/05/04 21:08:12.999999
    "%Y-%m-%d %H:%M:%S.%f",  # 2007-05-04 21:08:12.999999
    "%Y-%b-%d %H:%M:%S.%f",  # 2007-May-04 21:08:12.999999
    "%d-%b-%Y %H:%M:%S.%f",  # 04-May-2007 21:08:12.999999
    "%Y:%j:%H:%M:%S.%f",  # 2012:124:21:08:12.999999
]


def check_equal_time(t1: Time, t2: Time) -> bool:
    """
    Check if two time objects are equal within a nanosecond precision.

    :param t1: The first time object.
    :type t1: Time
    :param t2: The second time object.
    :type t2: Time
    :returns: True if the time difference is less than a nanosecond, False otherwise.
    :rtype: bool

    **Example:**

    .. code-block:: python

        >>> from astropy.time import Time
        >>> t1 = Time('2020-01-01T00:00:00.000000000')
        >>> t2 = Time('2020-01-01T00:00:00.000000001')
        >>> check_equal_time(t1, t2)  # Returns: True
    """
    if abs(t2 - t1) < 1 * u.nanosecond:
        return True
    return False


def regex_time(time_string: str, format: str) -> Optional[str]:
    """
    Match a time string against a format using regular expressions.

    :param time_string: The time string to match.
    :type time_string: str
    :param format: The time format string.
    :type format: str
    :returns: The time string if a match is found, otherwise None.
    :rtype: Optional[str]

    **Example:**

    .. code-block:: python

       >>>  regex_time('2020-01-01T00:00:00.000000', '%Y-%m-%dT%H:%M:%S.%f')
        # Returns: '2020-01-01T00:00:00.000000'
    """
    regex_pattern = format
    for key, value in TIME_REGEX.items():
        regex_pattern = regex_pattern.replace(key, value)
    # Обработка необязательного микросекундного компонента
    if '%f' in format:
        regex_pattern = regex_pattern.replace(r'(?P<microsecond>\d{1,6})', r'(?P<microsecond>\d{1,6})?')
    regex_pattern = '^' + regex_pattern + '$'  # Добавим начало и конец строки
    match = re.match(regex_pattern, time_string)
    if match is None:
        return None
    return time_string


def get_time_format(time_string: str) -> Optional[str]:
    """
    Determine the time format of a given time string.

    :param time_string: The time string to check.
    :type time_string: str
    :returns: The format string if a matching format is found, otherwise None.
    :rtype: Optional[str]

    :Example:

    .. code-block:: python

        >>> get_time_format('2020-01-01T00:00:00.000000')
        Returns: '%Y-%m-%dT%H:%M:%S.%f'
    """
    for time_format in COMMON_TIME_FORMATS:
        if regex_time(time_string, time_format) is not None:
            return time_format
    return None


@singledispatch
def create_time(time_string: Any, format: Optional[str] = None, **kwargs) -> Time:
    """
    Create an Astropy Time object from various time representations.

    :param time_string: The time data, which can be of various types.
    :type time_string: Any
    :param format: The time format string, if known.
    :type format: Optional[str]
    :param kwargs: Additional keyword arguments passed to ``astropy.time.Time``.
    :returns: An Astropy Time object.
    :rtype: Time

    **Example:**

    .. code-block:: python

        >>> create_time('2020-01-01T00:00:00.000000')
        # Returns: <Time object: scale='utc' format='isot' value=2020-01-01T00:00:00.000>
    """
    return Time(time_string, format=format, **kwargs)


@create_time.register(pd.Timestamp)
def create_time_pandasTimestamp(time_string: pd.Timestamp, **kwargs) -> Time:
    """
    Convert a pandas Timestamp to an Astropy Time object.

    :param time_string: A pandas Timestamp object.
    :type time_string: pd.Timestamp
    :param kwargs: Additional keyword arguments passed to ``astropy.time.Time``.
    :returns: An Astropy Time object.
    :rtype: Time

    **Example:**

    .. code-block:: python

        >>> ts = pd.Timestamp('2020-01-01T00:00:00')
        >>> create_time(ts)
        <Time object: scale='utc' format='datetime64' value=2020-01-01T00:00:00.000>
    """
    return Time(time_string.asm8)


@create_time.register(pd.Series)
def create_time_pandasSeries(time_string: pd.Series, **kwargs) -> Time:
    """
    Convert a pandas Series of timestamps to an Astropy Time object.

    :param time_string: A pandas Series of timestamp objects.
    :type time_string: pd.Series
    :param kwargs: Additional keyword arguments passed to ``astropy.time.Time``.
    :returns: An Astropy Time object.
    :rtype: Time

    **Example:**

    .. code-block:: python

        >>> ts_series = pd.Series([pd.Timestamp('2020-01-01T00:00:00')])
        >>> create_time(ts_series)
        <Time object: scale='utc' format='datetime64' value=[2020-01-01T00:00:00.000]>
    """
    return Time(time_string.tolist(), **kwargs)


@create_time.register(pd.DatetimeIndex)
def create_time_pandasDatetimeIndex(time_string: pd.DatetimeIndex, **kwargs) -> Time:
    """
    Convert a pandas DatetimeIndex to an Astropy Time object.

    :param time_string: A pandas DatetimeIndex object.
    :type time_string: pd.DatetimeIndex
    :param kwargs: Additional keyword arguments passed to ``astropy.time.Time``.
    :returns: An Astropy Time object.
    :rtype: Time

    **Example:**

    .. code-block:: python

        >>> dt_index = pd.DatetimeIndex(['2020-01-01T00:00:00'])
        >>> create_time(dt_index)
        <Time object: scale='utc' format='datetime64' value=[2020-01-01T00:00:00.000]>
    """
    return Time(time_string.tolist(), **kwargs)


@create_time.register(datetime)
def create_time_datetime(time_string: datetime, **kwargs) -> Time:
    """
    Convert a datetime object to an Astropy Time object.

    :param time_string: A datetime object.
    :type time_string: datetime
    :param kwargs: Additional keyword arguments passed to ``astropy.time.Time``.
    :returns: An Astropy Time object.
    :rtype: Time

    **Example:**

    .. code-block:: python

        >>> dt = datetime(2020, 1, 1)
        >>> create_time(dt)
        <Time object: scale='utc' format='datetime' value=2020-01-01T00:00:00.000>
    """
    return Time(time_string, **kwargs)


@create_time.register(date)
def create_time_date(time_string: date, **kwargs) -> Time:
    """
    Convert a date object to an Astropy Time object.

    :param time_string: A date object.
    :type time_string: date
    :param kwargs: Additional keyword arguments passed to ``astropy.time.Time``.
    :returns: An Astropy Time object.
    :rtype: Time

    **Example:**

    .. code-block:: python

        >>> d = date(2020, 1, 1)
        >>> create_time(d)
        <Time object: scale='utc' format='iso' value=2020-01-01T00:00:00.000>
    """
    return Time(time_string.isoformat(), **kwargs)


@create_time.register(np.datetime64)
def create_time_npdatetime64(time_string: np.datetime64, **kwargs) -> Time:
    """
    Convert a numpy datetime64 object to an Astropy Time object.

    :param time_string: A numpy datetime64 object.
    :type time_string: np.datetime64
    :param kwargs: Additional keyword arguments passed to ``astropy.time.Time``.
    :returns: An Astropy Time object.
    :rtype: Time

    **Example:**

    .. code-block:: python

        >>> dt64 = np.datetime64('2020-01-01T00:00:00')
        >>> create_time(dt64)
        <Time object: scale='utc' format='iso' value=2020-01-01T00:00:00.000>
    """
    return Time(str(time_string.astype('M8[ns]')), **kwargs)


@create_time.register(np.ndarray)
def create_time_npndarray(time_string: np.ndarray, **kwargs) -> Time:
    """
    Convert a numpy ndarray of datetime64 objects to an Astropy Time object.

    :param time_string: A numpy ndarray of datetime64 objects.
    :type time_string: np.ndarray
    :param kwargs: Additional keyword arguments passed to ``astropy.time.Time``.
    :returns: An Astropy Time object.
    :rtype: Time

    **Example:**

    .. code-block:: python

        >>> dt64_array = np.array(['2020-01-01T00:00:00'], dtype='datetime64')
        >>> create_time(dt64_array)
        <Time object: scale='utc' format='iso' value=[2020-01-01T00:00:00.000]>
    """
    if 'datetime64' in str(time_string.dtype):
        return Time([str(dt.astype('M8[ns]')) for dt in time_string], **kwargs)
    else:
        return create_time.dispatch(object)(time_string, **kwargs)


@create_time.register(astropy.time.Time)
def create_time_astropy(time_string: Time, **kwargs) -> Time:
    """
    Return the provided Astropy Time object.

    :param time_string: An Astropy Time object.
    :type time_string: Time
    :param kwargs: Additional keyword arguments passed to ``astropy.time.Time``.
    :returns: The provided Astropy Time object.
    :rtype: Time

    **Example:**

    .. code-block:: python

        >>> t = Time('2020-01-01T00:00:00')
        >>> create_time(t)
        <Time object: scale='utc' format='isot' value=2020-01-01T00:00:00.000>
    """
    return time_string


@create_time.register(list)
def create_time_list(time_list: list, format: Optional[str] = None, **kwargs) -> Time:
    """
    Convert a list of time strings to an Astropy Time object.

    :param time_list: A list of time strings.
    :type time_list: list
    :param format: The time format string, if known.
    :type format: Optional[str]
    :param kwargs: Additional keyword arguments passed to ``astropy.time.Time``.
    :returns: An Astropy Time object.
    :rtype: Time

    **Example:**

    .. code-block:: python

        >>> time_list = ['2020-01-01T00:00:00']
        >>> create_time(time_list)
        <Time object: scale='utc' format='isot' value=[2020-01-01T00:00:00.000]>
    """
    item = time_list[0]
    if isinstance(item, str) and format is None:
        time_format = get_time_format(item)
        return Time.strptime(time_list, time_format, **kwargs)

    return create_time_list.dispatch(object)(time_list, format, **kwargs)


@create_time.register(str)
def create_time_str(time_string: str, **kwargs) -> Time:
    """
    Convert a time string to an Astropy Time object using common time formats.

    :param time_string: A time string.
    :type time_string: str
    :param kwargs: Additional keyword arguments passed to ``astropy.time.Time``.
    :returns: An Astropy Time object.
    :rtype: Time

    **Example:**

    .. code-block:: python

        >>> create_time('2020-01-01T00:00:00')
        <Time object: scale='utc' format='isot' value=2020-01-01T00:00:00.000>
    """
    for time_format in COMMON_TIME_FORMATS:
        try:
            try:
                ts = regex_time(time_string, time_format)
            except TypeError:
                break
            if ts is None:
                continue
            t = Time.strptime(ts, time_format, **kwargs)
            return t
        except ValueError:
            pass

    return create_time.dispatch(object)(time_string, **kwargs)


def parse_time(time_string, format=None, **kwargs):
    if isinstance(time_string, str) and time_string == 'now':
        rt = Time.now()
    else:
        rt = create_time(time_string, format=format, **kwargs)
    return rt



import matplotlib.pyplot as plt
import matplotlib.colors as colors
import radiosunpy.visualization.colormaps as cm
from radiosunpy.client import RATANClient
from typing import Optional, Union
from astropy.io import fits
import numpy as np

def plot_ratan_fits_data(
                        pr_data: Union[str, fits.hdu.hdulist.HDUList],
                        is_calibrated: bool = False,
                        plot_V: Optional[bool] = True,
                        plot_I: Optional[bool] = True) -> None: 
    """
    Plot RATAN-600 FITS data for Stokes I and/or V parameters.
    
    :param pr_data: Path to FITS file, URL, or FITS HDUList with the RATAN-600 data.
    :type pr_data: Union[str, fits.hdu.hdulist.HDUList]
    :param is_calibrated: Indicates whether the data is already calibrated. Defaults to False.
    :type is_calibrated: bool
    :param plot_V: Whether to plot Stokes V data. Defaults to True.
    :type plot_V: Optional[bool]
    :param plot_I: Whether to plot Stokes I data. Defaults to True.
    :type plot_I: Optional[bool]
    :return: None
    :rtype: None
    :raises ValueError: If both plot_I and plot_V are set to False.
    """
    if not plot_I and not plot_V:
        raise ValueError("At least one of plot_I or plot_V must be True.")

    if isinstance(pr_data, str):
        ratan_client = RATANClient()
        raw, processed = ratan_client.process_fits_data(pr_data,
                                                    save_path=None,
                                                    save_with_original=False)
        ratan_file = processed if is_calibrated else raw
    elif isinstance(pr_data, fits.hdu.hdulist.HDUList):
        ratan_file = pr_data
        is_calibrated = True if ratan_file[0].data is None else False
    else:
        NotImplementedError

    SOLAR_R = ratan_file[0].header['SOLAR_R']
    CDELT1 = ratan_file[0].header['CDELT1']
    CRPIX1 = ratan_file[0].header['CRPIX1']
    DATE = ratan_file[0].header['DATE-OBS']

    if is_calibrated: 
        I, V, FREQ = ratan_file[1].data, ratan_file[2].data, ratan_file[3].data
    else: 
        I, V, FREQ = ratan_file[0].data[:, 0, :], ratan_file[0].data[:, 1, :], ratan_file[1].data['FREQ']

    x = np.linspace(
        - CRPIX1 * CDELT1,
        (I.shape[1] - CRPIX1) * CDELT1,
        num=I.shape[1]
    )

    if plot_I and not plot_V:
        min_data, max_data = np.min(I) * 1.1, np.max(I) * 1.1
    elif plot_V and not plot_I:
        min_data, max_data = np.min(V) * 1.1, np.max(V) * 1.1
    else:
        min_data, max_data = np.min(V) * 1.1, np.max(I) * 1.1

    title_start = 'RATAN-600 calibrated scan' if is_calibrated else 'RATAN-600 raw scan'
    modes = 'I-V' if plot_I and plot_V else 'I' if plot_I else 'V'
    title = f"{title_start} // {DATE} // {modes}"
    y_label = 'Spectral Flux Density, s.f.u' if is_calibrated else 'Antenna Temperature, K'

    fig, ax1 = plt.subplots(figsize=(8, 7))

    color_map = plt.get_cmap('std_gamma_2').reversed()
    norm = colors.Normalize(vmin=np.min(FREQ), vmax=np.max(FREQ))
    sm = plt.cm.ScalarMappable(cmap=color_map, norm=norm)
    sm.set_array([])

    size = len(FREQ)
    
    if plot_I:
        for i in range(size):
            color = color_map(i / (size - 1))
            ax1.plot(x, I[i], color=color)
        ax1.set_ylabel('(I) ' + y_label)

    ax1.set_facecolor('black')
    ax1.set_xlim(-1.3 * SOLAR_R, 1.3 * SOLAR_R)
    ax1.set_ylim(min_data, max_data)

    if plot_V:
        ax2 = ax1.twinx() if plot_I else ax1
        for i in range(len(FREQ)):
            color = color_map(i / (size - 1))
            ax2.plot(x, V[i], color=color)
        ax2.set_ylabel('(V) ' + y_label)
        if plot_I: 
          ax1.set_ylim(bottom=-0.2)
          ax2.set_ylim(top=0.5)

    ax1.set_xlabel('Distance from solar centre, arcsec') 

    plt.plot([0, 0], [min_data, max_data], 'y--')
    plt.plot([SOLAR_R, SOLAR_R], [min_data, max_data], 'y--')
    plt.plot([-SOLAR_R, -SOLAR_R], [min_data, max_data], 'y--')
    plt.axis([-1.3 * SOLAR_R, 1.3 * SOLAR_R, min_data, max_data])
    plt.colorbar(sm, ax=ax2 if plot_V else ax1, ticks=FREQ[::10], orientation="horizontal", label='Frequency, GHz')

    plt.title(title)

    fig.tight_layout()
    plt.show()



from pathlib import Path
import numpy as np 
from matplotlib.colors import LinearSegmentedColormap
import matplotlib

CMAPS_DIR = Path(__file__).absolute().parent.joinpath('colormap_files')

def cmap_from_file(file_name: str, 
                   cmap_name: str) -> LinearSegmentedColormap: 
    """
    Create a colormap from a CSV file containing RGB values.

    :param file_name: Name of the CSV file (without extension) containing RGB data.
    :type file_name: str
    :param cmap_name: The name to assign to the generated colormap.
    :type cmap_name: str
    :return: A LinearSegmentedColormap object based on the RGB data.
    :rtype: matplotlib.colors.LinearSegmentedColormap
    """
    color_data = np.loadtxt(CMAPS_DIR.joinpath(file_name + '.csv'), delimiter=',')
    r, g, b = color_data[:, 0], color_data[:, 1], color_data[:, 2]
    i = np.linspace(0, 1, r.size)
    color_dict = {
        name: list(zip(i, el / 255.0, el / 255.0))
        for el, name in [(r, 'red'), (g, 'green'), (b, 'blue')]
    }
    return LinearSegmentedColormap(cmap_name, color_dict)

def std_gamma_2():
    return cmap_from_file('std_gamma_2', 'std_gamma_2.csv')

colormaps_list = {
    'std_gamma_2': std_gamma_2()
}

for name, colormap in colormaps_list.items():
    matplotlib.colormaps.register(colormap, name=name)


import numpy as np
import pywt
from typing import Union, Any
from scipy import ndimage

from numpy import signedinteger


from pathlib import Path

def get_project_root() -> Path:
    return Path(__file__).parent.parent.parent

def SunSolidAngle(R: float) -> float:
    """
     Calculate the solid angle subtended by the Sun as seen from Earth.

     :param R: Radius of the Sun's disc in arcseconds.
     :type R: float
     :return: Solid angle in degrees.
     :rtype: float

     :Example:

     >>> SunSolidAngle(960)
     6.805218475785968e-05
     """
    R_deg = (R / 3600)
    return np.pi * (np.pi * R_deg / 180) ** 2


def gauss2d(x: Union[np.ndarray, list], y: Union[np.ndarray, list],
            amplitude_x: float, amplitude_y: float,
            mean_x: float, mean_y: float,
            sigma_x: float, sigma_y: float) -> np.ndarray:
    """
        Generate a 2D Gaussian distribution.

        :param x: Array of x coordinates.
        :type x: numpy.ndarray or list of floats
        :param y: Array of y coordinates.
        :type y: numpy.ndarray or list of floats
        :param amplitude_x: Amplitude along the x-axis.
        :type amplitude_x: float
        :param amplitude_y: Amplitude along the y-axis.
        :type amplitude_y: float
        :param mean_x: Mean of the Gaussian along the x-axis.
        :type mean_x: float
        :param mean_y: Mean of the Gaussian along the y-axis.
        :type mean_y: float
        :param sigma_x: Standard deviation along the x-axis.
        :type sigma_x: float
        :param sigma_y: Standard deviation along the y-axis.
        :type sigma_y: float
        :return: 2D Gaussian array.
        :rtype: numpy.ndarray

        :Example:

        >>> x = np.linspace(-1, 1, 100)
        >>> y = np.linspace(-1, 1, 100)
        >>> gauss2d(x, y, 1, 1, 0, 0, 0.1, 0.1).shape
        (100, 100)
        """
    x, y = np.meshgrid(x, y)
    g = amplitude_x * amplitude_y * np.exp(
        -((x - mean_x) ** 2 / (2 * sigma_x ** 2) + (y - mean_y) ** 2 / (2 * sigma_y ** 2)))
    return g


def gauss1d(x: np.ndarray, amplitude: float,
            mean: float, sigma: float) -> np.ndarray:
    """
        Generate a 1D Gaussian distribution.

        :param x: Array of x coordinates.
        :type x: numpy.ndarray
        :param amplitude: Amplitude of the Gaussian.
        :type amplitude: float
        :param mean: Mean of the Gaussian.
        :type mean: float
        :param sigma: Standard deviation of the Gaussian.
        :type sigma: float
        :return: 1D Gaussian array.
        :rtype: numpy.ndarray

        :Example:

        >>> x = np.linspace(-1, 1, 100)
        >>> gauss1d(x, 1, 0, 0.1).shape
        (100,)
        """
    return amplitude * np.exp(-((x - mean) ** 2) / (2 * sigma ** 2))


def gaussian_mixture(params: list[float],
                     x: np.ndarray,
                     y: np.ndarray) -> np.ndarray:
    """
        Model data as a mixture of 1D Gaussians and return the residuals.

        :param params: List of Gaussian parameters [amplitude, mean, sigma, amplitude, mean, sigma, ....].
        :type params: list of floats
        :param x: Array of x coordinates.
        :type x: numpy.ndarray
        :param y: Array of observed data values.
        :type y: numpy.ndarray
        :return: Residuals between the model and the observed data.
        :rtype: numpy.ndarray

        :Example:

        >>> params = [1, 0, 0.1, 0.5, 0.5, 0.1]
        >>> x = np.linspace(-1, 1, 100)
        >>> y = np.ones(100)
        >>> gaussian_mixture(params, x, y).shape
        (100,)
        """
    model = np.zeros_like(x)
    for i in range(0, len(params), 3):
        amplitude = params[i]
        mean = params[i + 1]
        stddev = params[i + 2]
        model += gauss1d(x, amplitude, mean, stddev)
    return y - model


def create_rectangle(size: int,
                     width: float,
                     height: float) -> np.ndarray:
    """
    Create a 2D rectangle of the given width and height within a square array.

    :param size: Size of the square array (size x size).
    :type size: int
    :param width: Width of the rectangle.
    :type width: float
    :param height: Height of the rectangle.
    :type height: float
    :return: 2D array with the rectangle.
    :rtype: numpy.ndarray

    :Example:

    >>> create_rectangle(100, 50, 20).shape
    (100, 100)
    """
    x = np.linspace(-size // 2, size // 2, size)
    y = np.linspace(-size // 2, size // 2, size)
    x, y = np.meshgrid(x, y)
    rectangle = np.zeros((size, size))
    mask = (abs(x) <= width / 2) & (abs(y) <= height / 2)
    rectangle[mask] = 1
    return rectangle


def create_sun_model(size: int, radius: float) -> np.ndarray:
    """
    Create a 2D circular model representing the Sun's disc.

    :param size: Size of the square array (size x size).
    :type size: int
    :param radius: Radius of the Sun's disc in pixels.
    :type radius: int
    :return: 2D array with the circular Sun model.
    :rtype: numpy.ndarray

    :Example:

    >>> create_sun_model(100, 20).shape
    (100, 100)
    """
    y, x = np.ogrid[-size // 2:size // 2, -size // 2:size // 2]
    mask = x ** 2 + y ** 2 <= radius ** 2
    sun_model = np.zeros((size, size))
    sun_model[mask] = 1
    return sun_model


def calculate_area(image: np.ndarray) -> float:
    """
    Calculate the area under a 1D curve or a 2D surface.

    :param image: Array representing the image or curve.
    :type image: numpy.ndarray
    :return: The calculated area.
    :rtype: float

    :Example:

    >>> calculate_area(np.ones(100))
    99.0
    """
    return float(np.trapz(image))


def bwhm_to_sigma(bwhm: float) -> float:
    """
    Convert beam width at half maximum (BWHM) to the standard deviation (sigma).

    :param bwhm: Beam width at half maximum.
    :type bwhm: float
    :return: Corresponding standard deviation (sigma).
    :rtype: float

    :Example:

    >>> bwhm_to_sigma(2.355) # doctest: +ELLIPSIS
    0.7071608181730225
    """
    fwhm = np.sqrt(1 / 2) * bwhm
    return float(fwhm / (2 * np.sqrt(2 * np.log(2))))


def flip_and_concat(values: np.ndarray,
                    flip_values: bool = False) -> np.ndarray:
    """
    Flip and concatenate an array.

    :param values: Array of values to be flipped and concatenated.
    :type values: numpy.ndarray
    :param flip_values: If True, flips the sign of the flipped values.
    :type flip_values: bool
    :return: Concatenated array.
    :rtype: numpy.ndarray

    :Example:

    >>> flip_and_concat(np.array([1, 2, 3]), True)
    array([-3, -2,  1,  2,  3])
    """
    flipped = -values[::-1][:-1] if flip_values else values[::-1][:-1]
    return np.concatenate((flipped, values))


def error(scale_factor: float,
          experimental_data: np.ndarray,
          theoretical_data: np.ndarray) -> float:
    """
    Calculate the squared error between experimental and theoretical data.

    :param scale_factor: Scaling factor applied to experimental data.
    :type scale_factor: float
    :param experimental_data: Array of experimental data values.
    :type experimental_data: numpy.ndarray
    :param theoretical_data: Array of theoretical data values.
    :type theoretical_data: numpy.ndarray
    :return: The squared error.
    :rtype: float

    :Example:

    >>> error(1, np.ones(100), np.zeros(100)) # doctest: +ELLIPSIS
    100.0
    """
    return float(np.sum((scale_factor * experimental_data - theoretical_data) ** 2))


def wavelet_denoise(data: np.ndarray, wavelet, level):
    """
    Denoise a signal using wavelet decomposition.

    :param data: Array of data values to be denoised.
    :type data: numpy.ndarray
    :param wavelet: Wavelet type to be used for decomposition.
    :type wavelet: str
    :param level: Level of decomposition.
    :type level: int
    :return: Denoised data.
    :rtype: numpy.ndarray

    :Example:

    >>> data = np.random.normal(0, 1, 100)
    >>> wavelet_denoise(data, 'sym6', 2).shape
    (100,)
    """
    N = 2 ** np.ceil(np.log2(len(data))).astype(int)
    data_padded = np.pad(data, (0, N - len(data)), 'constant', constant_values=(0, 0))
    coeff = pywt.wavedec(data_padded, wavelet, mode="sym")
    sigma = np.median(np.abs(coeff[-level] - np.median(coeff[-level]))) / 0.6745
    uthresh = sigma * np.sqrt(2 * np.log(len(data_padded)))
    coeff[1:] = (pywt.threshold(i, value=uthresh, mode="soft") for i in coeff[1:])
    return pywt.waverec(coeff, wavelet, mode="sym")[:len(data)]


def clipped_zoom(img, zoom_factor, **kwargs):
    h, w = img.shape[:2]

    # For multichannel images we don't want to apply the zoom factor to the RGB
    # dimension, so instead we create a tuple of zoom factors, one per array
    # dimension, with 1's for any trailing dimensions after the width and height.
    zoom_tuple = (zoom_factor,) * 2 + (1,) * (img.ndim - 2)

    # Zooming out
    if zoom_factor < 1:

        # Bounding box of the zoomed-out image within the output array
        zh = int(np.round(h * zoom_factor))
        zw = int(np.round(w * zoom_factor))
        top = (h - zh) // 2
        left = (w - zw) // 2

        # Zero-padding
        out = np.zeros_like(img)
        out[top:top + zh, left:left + zw] = ndimage.zoom(img, zoom_tuple, **kwargs)

    # Zooming in
    elif zoom_factor > 1:

        # Bounding box of the zoomed-in region within the input array
        zh = int(np.round(h / zoom_factor))
        zw = int(np.round(w / zoom_factor))
        top = (h - zh) // 2
        left = (w - zw) // 2

        out = ndimage.zoom(img[top:top + zh, left:left + zw], zoom_tuple, **kwargs)

        # `out` might still be slightly larger than `img` due to rounding, so
        # trim off any extra pixels at the edges
        trim_top = ((out.shape[0] - h) // 2)
        trim_left = ((out.shape[1] - w) // 2)
        out = out[trim_top:trim_top + h, trim_left:trim_left + w]

    # If zoom_factor == 1, just return the input array
    else:
        out = img
    return out



import re
from datetime import datetime
from ftplib import FTP
from typing import List, Optional, Callable, Any
from urllib.parse import urlsplit
import requests
from astropy.time import Time
from dateutil.relativedelta import relativedelta

from radiosunpy.time import TimeRange

TIME_REGEX = {'%Y': r'\d{4}', '%y': r'\d{2}',
              '%b': '[A-Z][a-z]{2}', '%m': r'\d{2}',
              '%d': r'\d{2}', '%j': r'\d{3}',
              '%H': r'\d{2}',
              '%M': r'\d{2}',
              '%S': r'\d{2}'}

__all__ = ['Scrapper']

class Scrapper:
    def __init__(
            self,
            baseurl: str,
            regex_pattern: Optional[str] = None,
            condition: Optional[Callable[[str, str, str], str]] = None,
            filter: Optional[Callable[[str], bool]] = None,
            **kwargs: Any
    ) -> None:
        """
        Initialize the Scrapper object with base URL and optional parameters.

        :param baseurl: The base URL to scrape data from.
        :param regex_pattern: Optional regex pattern to match filenames.
        :param condition: Optional callable to generate dates based on extracted data.
        :param filter: Optional filter to apply to the extracted data.
        """

        self.baseurl = baseurl
        self.domain = f"{urlsplit(self.baseurl).scheme}://{urlsplit(self.baseurl).netloc}/"
        self.filter = filter
        self.regex_pattern = regex_pattern
        self.condition = condition

    @staticmethod
    def smallest_significant_pattern(pattern: str) -> Optional[relativedelta]:
        """
        Determine the smallest significant pattern (e.g., seconds, minutes, days) in the given pattern.
        Some of them are here: https://fits.gsfc.nasa.gov/iso-time.html

        :param pattern: The pattern string.
        :return: The smallest significant `relativedelta` object, or None if not found.
        """
        try:
            if any(second in pattern for second in ['%S']):
                return relativedelta(seconds=1)
            elif any(minute in pattern for minute in ['%M']):
                return relativedelta(minutes=1)
            elif any(hour in pattern for hour in ['%H']):
                return relativedelta(hours=1)
            elif any(day in pattern for day in ['%d', '%j']):
                return relativedelta(days=1)
            elif any(month in pattern for month in ['%m', '%b']):
                return relativedelta(months=1)
            if any(year in pattern for year in ['%y', '%Y']):
                return relativedelta(years=1)
            else:
                return None
        except Exception:
            raise

    @staticmethod
    def floor_datetime(date: Time, timestep: relativedelta) -> datetime:
        """
        Floor the given datetime to the nearest significant time unit.

        :param date: The `Time` object to floor.
        :param timestep: The `relativedelta` object representing the smallest significant time unit.
        :return: The floored `datetime` object.
        """
        date = date.to_datetime()
        if timestep.years > 0:
            return date.replace(month=1, day=1, hour=0, minute=0, second=0, microsecond=0)
        elif timestep.months > 0:
            return date.replace(day=1, hour=0, minute=0, second=0, microsecond=0)
        elif timestep.days > 0:
            return date.replace(hour=0, minute=0, second=0, microsecond=0)
        elif timestep.hours > 0:
            return date.replace(minute=0, second=0, microsecond=0)
        elif timestep.minutes > 0:
            return date.replace(second=0, microsecond=0)
        return date

    def range(self, timerange: TimeRange) -> List[str]:
        """
        Generate a list of directories within the time range based on the smallest significant pattern.

       :param timerange: The `TimeRange` object representing the time range.
       :return: A list of directory paths.
       """
        filepath_pattern = '/'.join(self.baseurl.split('/')[:-1]) + '/'
        smallest_step = self.smallest_significant_pattern(filepath_pattern)
        if smallest_step is None:
            return [filepath_pattern]
        directories = []
        current_date = self.floor_datetime(timerange.start, smallest_step)
        end_date = self.floor_datetime(timerange.end, smallest_step) + smallest_step
        while current_date < end_date:
            directories.append(current_date.strftime(filepath_pattern))
            current_date += smallest_step
        return directories

    def extract_date_from_url(self, url):
        """
        Extract date from a given URL based on the base URL's pattern.

        :param url: The URL string.
        :return: The extracted `Time` object.
        """
        def url2list(text):
            return re.sub(r'\.|_', '/', text).split('/')

        pattern_parts = url2list(self.baseurl)
        url_parts = url2list(url)
        time_formats = ['%Y', '%y', '%b', '%B', '%m', '%d', '%j',
                        '%H', '%M', '%S']
        extracted_date, extracted_pattern = [], []
        for p_elem, u_elem in zip(pattern_parts, url_parts):
            present_formats = [x for x in time_formats if x in p_elem]
            part_to_remove = re.split('%.', p_elem)

            for candidate_to_remove in part_to_remove:
                if candidate_to_remove:
                    u_elem = u_elem.replace(candidate_to_remove, '', 1)
                    p_elem = p_elem.replace(candidate_to_remove, '', 1)

            extracted_date.append(u_elem)
            extracted_pattern.append(p_elem)
            time_formats = [fmt for fmt in time_formats if fmt not in present_formats]

        # Find the index of the fullest pattern
        fullest_pattern_index = extracted_pattern.index(max(extracted_pattern, key=len))
        # Find the corresponding date part
        date_part = extracted_date[fullest_pattern_index]
        return Time.strptime(date_part, extracted_pattern[fullest_pattern_index])

    def valid_date_from_url(self, url: str) -> bool:
        """
        Validate if a given URL's date matches the expected pattern from the base URL.

        :param url: The URL string to validate.
        :return: True if the URL's date matches the pattern, False otherwise.
        """
        pattern = self.baseurl
        # Replace datetime formats in the pattern string with their corresponding regex patterns
        for time_format, regex in TIME_REGEX.items():
            pattern = pattern.replace(time_format, regex)
        # Create a regex pattern object
        pattern_obj = re.compile(pattern)
        # Check if the URL matches the pattern
        return pattern_obj.fullmatch(url) is not None

    def check_date_in_timerange_from_url(self,
                                         url: str,
                                         timerange: TimeRange) -> bool:
        """
        Check if the date extracted from a URL is within the given time range.

        :param url: The URL string.
        :param timerange: The `TimeRange` object representing the time range.
        :return: True if the date is within the range, False otherwise.
        """
        file_date = self.extract_date_from_url(url).to_datetime()
        #smallest_pattern = self.smallest_significant_pattern(self.baseurl)
        file_range = TimeRange(file_date, file_date)
        return timerange.have_intersection(file_range)

    def check_date_in_timerange_from_file_date(self,
                                               file_date: str,
                                               timerange: TimeRange) -> bool:
        """
        Check if a given file date is within the specified time range.

        :param file_date: The file date as a string (format: "%Y-%m-%d").
        :param timerange: The `TimeRange` object representing the time range.
        :return: True if the date is within the range, False otherwise.
        """
        file_date = datetime.strptime(file_date, "%Y-%m-%d")
        #smallest_pattern = self.smallest_significant_pattern(self.baseurl)
        file_range = TimeRange(file_date, file_date)
        return timerange.have_intersection(file_range)

    def ftpfiles(self, timerange: TimeRange) -> List[str]:
        """
        Retrieve a list of files from an FTP server within the specified time range.

        :param timerange: The `TimeRange` object representing the time range.
        :return: A list of file URLs.
        """
        directories = self.range(timerange)
        file_urls = []
        ftpurl = urlsplit(directories[0]).netloc
        with FTP(ftpurl, user="anonymous", passwd="soleil@package") as ftp:
            for current_directory in directories:
                try:
                    ftp.cwd(urlsplit(current_directory).path)
                except Exception as e:
                    print(f'FTP CWD tried: {e}')
                    continue
                for file_name in ftp.nlst():
                    file_path = current_directory + file_name
                    if self.check_date_in_timerange_from_url(file_path, timerange):
                        file_urls.append(file_path)
        return file_urls

    def httpfiles(self, timerange: TimeRange) -> List[str]:
        """
        Retrieve a list of files from an HTTP server within the specified time range.

        :param timerange: The `TimeRange` object representing the time range.
        :return: A list of file URLs.
        """
        directories = self.range(timerange)
        file_urls = []
        for current_directory in directories:
            directory_parts = current_directory.split('/')
            year = directory_parts[-3]
            month = directory_parts[-2]
            try:
                page = requests.get(current_directory)
                page.raise_for_status()
            except (requests.exceptions.RequestException, ConnectionResetError) as err:
                continue
            for match in re.findall(fr'href="{self.regex_pattern}"', page.text):
                relative_path, date_text = match
                date = self.condition(year, month,
                                      date_text) if self.condition else f'{date_text[:-4]}-{date_text[-4:-2]}-{date_text[-2:]}'
                url = current_directory + relative_path
                if self.check_date_in_timerange_from_file_date(date, timerange):
                    file_urls.append(url)
        return file_urls

    def form_fileslist(self, timerange: TimeRange) -> List[str]:
        """
        Retrieve a list of files from an HTTP or FTP server within the specified time range.

        :param timerange: The `TimeRange` object representing the time range.
        :type timerange: TimeRange
        :return: A list of file URLs.

        :Example: usage example based on SWPC Solar Region Summary (FTP server)

        >>> base_url_SRS = r'ftp://ftp.ngdc.noaa.gov/STP/swpc_products/daily_reports/solar_region_summaries/%Y/%m/%Y%m%dSRS.txt'
        >>> scraper = Scrapper(base_url_SRS)
        >>> t = TimeRange('2021-10-12', '2021-10-12')
        >>> print(t)
        (2021-10-12 00:00:00, 2021-10-12 00:00:00)

        >>> for url in scraper.form_fileslist(t):
        >>>     print(f'SRS url: {url}')
        SRS url: ftp://ftp.ngdc.noaa.gov/STP/swpc_products/daily_reports/solar_region_summaries/2021/10/20211012SRS.txt

        :Example: usage example based on RATAN (HTTP server)

        >>> if int(year) < 2010 or (int(year) == 2010 and int(month) < 5):
        >>>     return f'{year[:2]}{date_match[:-4]}-{date_match[-4:-2]}-{date_match[-2:]}'
        >>> else:
        >>>    f'{date_match[:-4]}-{date_match[-4:-2]}-{date_match[-2:]}'
        >>> base_url_RATAN = 'http://spbf.sao.ru/data/ratan/%Y/%m/%Y%m%d_%H%M%S_sun+0_out.fits'
        >>> regex_pattern_RATAN = '((\d{6,8})[^0-9].*[^0-9]0_out.fits)'
        >>> scraper = Scrapper(base_url_RATAN, regex_pattern=regex_pattern_RATAN, condition=build_date)
        >>> t = TimeRange('2010-01-13', '2010-01-13')
        >>> for url in scraper.form_fileslist(t):
        >>>     print(f'RATAN url: {url}')
        RATAN url: http://spbf.sao.ru/data/ratan/2010/01/100113sun0_out.fits
        """
        # SWPC SRS, for example
        if urlsplit(self.baseurl).scheme == 'ftp':
            return self.ftpfiles(timerange)
        # RATAN, for example
        if urlsplit(self.baseurl).scheme in ['http', 'https']:
            return self.httpfiles(timerange)



from datetime import timedelta
from datetime import datetime
import astropy.units as u
from typing import List, Optional, Union, Tuple
from astropy.time import Time, TimeDelta
from radiosunpy.time.time import parse_time, check_equal_time

TIME_FORMAT = "%Y-%m-%d %H:%M:%S"

__all__ = ['TimeRange']

class TimeRange:
    """
    A class to represent a time range, providing various utilities for time manipulation.

    **Attributes**

    - **start** (`Time`): The start time of the range.
    - **end** (`Time`): The end time of the range.
    - **delta** (`TimeDelta`): The duration of the time range.
    - **days** (`float`): The duration of the time range in days.
    - **hours** (`float`): The duration of the time range in hours.
    - **minutes** (`float`): The duration of the time range in minutes.
    - **seconds** (`float`): The duration of the time range in seconds.

    **Methods**

    - **__init__(a, b=None, format=None)**: Initializes a `TimeRange` object.
    - **__eq__(other)**: Checks if two `TimeRange` objects are equal.
    - **__ne__(other)**: Checks if two `TimeRange` objects are not equal.
    - **__contains__(time)**: Checks if a given time is within the time range.
    - **__repr__()**: Returns a string representation of the `TimeRange` object.
    - **__str__()**: Returns a simple string representation of the start and end times.
    - **have_intersection(other)**: Checks if the time range intersects with another `TimeRange`.
    - **get_dates(filter=None)**: Returns a list of dates within the time range, optionally excluding certain dates.
    - **moving_window(window_size, window_period)**: Returns a list of `TimeRange` objects representing moving windows within the original time range.
    - **equal_split(n_splits=2)**: Splits the time range into equal subranges.
    - **shift_forward(delta=None)**: Shifts the time range forward by a specified delta or its own duration.
    - **shift_backward(delta=None)**: Shifts the time range backward by a specified delta or its own duration.
    - **extend_range(start_delta, end_delta)**: Extends the start and end times of the time range by the specified deltas.
    """
    def __init__(self, a: Union['TimeRange', List[Union[str, Time]], Tuple[Union[str, Time], Union[str, Time]], str, Time],
                 b: Optional[Union[str, Time, timedelta, TimeDelta]] = None,
                 format: Optional[str] = None):
        self._start_time: Optional[Time] = None
        self._end_time: Optional[Time] = None

        # if TimeRange passed 
        if isinstance(a, TimeRange):
            self.__dict__ = a.__dict__.copy()
            return

            # if b is None and a is array-like or tuple.
        # for example, data and delta to the end of timerange
        if b is None:
            x = parse_time(a[0], format=format)
            if len(a) != 2:
                raise ValueError('"a" must have two elements')
            else:
                y = a[1]
        else:
            x = parse_time(a, format=format)
            y = b

            # if y is timedelta
        if isinstance(y, timedelta):
            y = TimeDelta(y, format='datetime')

        # create TimeRange in case of (start_date, delta)
        if isinstance(y, TimeDelta):
            # positive delta 
            if y.jd >= 0:
                self._start_time = x
                self._end_time = x + y
            else:
                self._start_time = x + y
                self._end_time = x
            return

        # otherwise, b is something date-like
        y = parse_time(y, format=format)
        if isinstance(y, Time):
            if x < y:
                self._start_time = x
                self._end_time = y
            else:
                self._start_time = y
                self._end_time = x

    @property
    def start(self):
        return self._start_time

    @property
    def end(self):
        return self._end_time

    @property
    def delta(self):
        return self._end_time - self._start_time

    @property
    def days(self):
        return self.delta.to('day').value

    @property
    def hours(self):
        return self.delta.to('hour').value

    @property
    def minutes(self):
        return self.delta.to('minute').value

    @property
    def seconds(self):
        return self.delta.to('second').value

    def __eq__(self, other: 'TimeRange') -> bool:
        if isinstance(other, TimeRange):
            return check_equal_time(self.start, other.start) and check_equal_time(self.end, other.end)
        return NotImplemented

    def __ne__(self, other: 'TimeRange') -> bool:
        if isinstance(other, TimeRange):
            return not (check_equal_time(self.start, other.start) and check_equal_time(self.end, other.end))
        return NotImplemented

    def __contains__(self, time: Union[str, Time]) -> bool:
        time_to_check = parse_time(time)
        return time_to_check >= self.start and time_to_check <= self.end

    def __repr__(self) -> str:
        start_time = self.start.strftime(TIME_FORMAT)
        end_time = self.end.strftime(TIME_FORMAT)
        full_name = f'{self.__class__.__module__}.{self.__class__.__name__}'
        return (
                f'<{full_name} object at {hex(id(self))}>' +
                '\nStart:'.ljust(12) + start_time +
                '\nEnd:'.ljust(12) + end_time +
                '\nDuration:'.ljust(
                    12) + f'{str(self.days)} days | {str(self.hours)} hours | {str(self.minutes)} minutes | {str(self.seconds)} seconds'
        )

    def __str__(self) -> str:
        start_time = self.start.strftime(TIME_FORMAT)
        end_time = self.end.strftime(TIME_FORMAT)
        return (
            f'({start_time}, {end_time})'
        )

    def have_intersection(self, other: 'TimeRange') -> bool:
        """
        Checks if the time range intersects with another TimeRange.

       :param other: Another TimeRange object to compare.
       :type other: TimeRange
       :returns: True if the ranges intersect, False otherwise.
       :rtype: bool``

        """

        intersection_lower = max(self.start, other.start)
        intersection_upper = min(self.end, other.end)
        return intersection_lower <= intersection_upper

    def get_dates(self, filter: Optional[List[datetime]] = None) -> List[datetime]:
        """
        Generate a list of dates within the time range defined by the instance.

        :param filter: A list of dates to be excluded from the result. If provided, only dates not in this list will be included.
        :type filter: list of datetime-like objects or None
        :return: A list of dates from start to end of the time range, with optional exclusion.
        :rtype: list of datetime-like objects
        """
        delta = self.end.to_datetime().date() - self.start.to_datetime().date()
        t_format = "%Y-%m-%d"
        dates_list = [parse_time(self.start.strftime(t_format)) + TimeDelta(i * u.day)
                      for i in range(delta.days + 1)]
        # filter is a list of dates to be excluded, maybe should add typings
        if filter:
            dates_list = [date for date in dates_list if date not in parse_time(filter)]
        return dates_list

    def moving_window(self, window_size: Union[TimeDelta, int], window_period: Union[TimeDelta, int]) -> List['TimeRange']:
        """
        Generate a list of time ranges using a moving window approach.

        :param window_size: The duration of each time window.
        :type window_size: TimeDelta or int
        :param window_period: The period to shift the window after each iteration.
        :type window_period: TimeDelta or int
        :return: A list of `TimeRange` objects representing the moving windows.
        :rtype: list of TimeRange
        """
        if not isinstance(window_size, TimeDelta):
            window_size = TimeDelta(window_size)
        if not isinstance(window_period, TimeDelta):
            window_period = TimeDelta(window_period)

        window_number = 1
        times = [TimeRange(self.start, self.start + window_size)]

        while times[-1].end < self.end:
            times.append(
                TimeRange(
                    self.start + window_number * window_period,
                    self.start + window_number * window_period + window_size,
                )
            )
            print(times)
            window_number += 1
        return times

    def equal_split(self, n_splits: int = 2) -> List['TimeRange']:
        """
        Split the time range into equal subranges.

        :param n_splits: The number of subranges to divide the time range into. Must be greater than or equal to 1.
        :type n_splits: int
        :raises ValueError: If `n_splits` is less than or equal to 0.
        :return: A list of `TimeRange` objects representing the equal subranges.
        :rtype: list of TimeRange
        """
        if n_splits <= 0:
            raise ValueError('n must be greater or equal than 1')
        subranges = []
        prev_time = self.start
        next_time = None
        for _ in range(n_splits):
            next_time = prev_time + self.delta / n_splits
            next_range = TimeRange(prev_time, next_time)
            subranges.append(next_range)
            prev_time = next_time
        return subranges

    def shift_forward(self, delta: Optional[TimeDelta] = None) -> 'self':
        """
        Shift the entire time range forward by a specified duration.

        :param delta: The duration by which to shift the time range. If not provided, the entire duration of the time range is used.
        :type delta: TimeDelta or None
        :return: The instance with the updated time range.
        :rtype: self
        """
        delta = delta if delta else self.delta
        self._start_time += delta
        self._end_time += delta
        return self

    def shift_backward(self, delta: Optional[TimeDelta] = None) -> 'self':
        """
        Shift the entire time range backward by a specified duration.

        :param delta: The duration by which to shift the time range. If not provided, the entire duration of the time range is used.
        :type delta: TimeDelta or None
        :return: The instance with the updated time range.
        :rtype: self
        """
        delta = delta if delta else self.delta
        self._start_time -= delta
        self._end_time -= delta
        return self

    def extend_range(self, start_delta: TimeDelta, end_delta: TimeDelta) -> 'self':
        """
        Extend the time range by modifying the start and end times.

        :param start_delta: The amount of time to add to the start of the range.
        :type start_delta: TimeDelta
        :param end_delta: The amount of time to add to the end of the range.
        :type end_delta: TimeDelta
        :return: The instance with the extended time range.
        :rtype: self
        """
        self._start_time += start_delta
        self._end_time += end_delta
        return self



from abc import ABCMeta, abstractmethod
import pathlib
from typing import List, Tuple, Any, Optional
from pathlib import Path
from urllib.request import urlopen
from collections import OrderedDict
from datetime import datetime
import astropy.io.ascii
from astropy.table import Column, MaskedColumn, vstack, Table, join
from astropy.io import fits
import pandas as pd
from scipy.interpolate import interp1d
from scipy.integrate import trapezoid, cumulative_trapezoid
from scipy.optimize import minimize
from scipy.optimize import curve_fit
from scipy.optimize import leastsq
from scipy.signal import fftconvolve
from scipy.signal import find_peaks
import numpy as np
from functools import lru_cache
import re
from radiosunpy.scrapper import Scrapper
from radiosunpy.time import TimeRange
from radiosunpy.utils import *

__all__ = ['RATANClient', 'SRSClient']


class BaseClient(metaclass=ABCMeta):
    @abstractmethod
    def acquire_data(self):
        pass

    @abstractmethod
    def form_data(self):
        pass

    @abstractmethod
    def get_data(self):
        pass


class SRSClient(BaseClient):
    base_url = 'ftp://ftp.ngdc.noaa.gov/STP/swpc_products/daily_reports/solar_region_summaries/%Y/%m/%Y%m%dSRS.txt'

    def extract_lines(self, content: str) -> object:
        """

        :type content: str with the source data by link
        """
        section, final_section = [], []

        for i, line in enumerate(content):
            if re.match(r'^(I\.|IA\.|II\.)', line):
                section.append(i)
            if re.match(
                    r'^(III|COMMENT|EFFECTIVE 2 OCT 2000|PLAIN|This message is for users of the NOAA/SEC Space|NNN)',
                    line, re.IGNORECASE):
                final_section.append(i)

        if final_section and final_section[0] > section[-1]:
            section.append(final_section[0])
        header = content[:section[0]] + [content[s] for s in section]
        for line in section:
            content[line] = '# ' + content[line]

        table1 = content[section[0]:section[1]]
        table1[1] = re.sub(r'Mag\s*Type', r'Magtype', table1[1], flags=re.IGNORECASE)
        table2 = content[section[1]:section[2]]
        if len(section) > 3:
            table3 = content[section[2]:section[3]]
            extra_lines = content[section[3]:]
        else:
            table3 = content[section[2]:]
            extra_lines = None
        data = [table1, table2, table3]
        for i, table in enumerate(data):
            if len(table) > 2 and table[2].strip().title() == 'None':
                del table[2]
        return header, data, extra_lines

    def proccess_lines(self, date, key, lines):
        column_mapping = {
            'Nmbr': 'Number',
            'Location': 'Location',
            'Lo': 'Carrington Longitude',
            'Area': 'Area',
            'Z': 'Z',
            'Ll': 'Longitudinal Extent',
            'Nn': 'Number of Sunspots',
            'Magtype': 'Mag Type',
            'Lat': 'Lat'
        }

        column_types = {
            'Number': np.dtype('i4'),
            'Location': np.dtype('U6'),
            'Carrington Longitude': np.dtype('i8'),
            'Area': np.dtype('i8'),
            'Z': np.dtype('U3'),
            'Longitudinal Extent': np.dtype('i8'),
            'Number of Sunspots': np.dtype('i8'),
            'Mag Type': np.dtype('S4'),
            'Lat': np.dtype('i8'),
        }

        if lines:
            raw_data = astropy.io.ascii.read(lines)
            column_names = list(raw_data.columns)
            raw_data.rename_columns(
                column_names, new_names=[column_mapping[col.title()] for col in column_names]
            )

            if len(raw_data) == 0:
                for c in raw_data.itercols():
                    c.dtype = column_types[c._name]
                raw_data.add_column(Column(data=None, name="ID", dtype=('S2')), index=0)
                raw_data.add_column(Column(data=None, name="Date", dtype=('S10')), index=0)
            else:
                raw_data.add_column(Column(data=[key] * len(raw_data), name="ID"), index=0)
                raw_data.add_column(Column(data=[date] * len(raw_data), name="Date"), index=0)
            return raw_data
        return None

    def parse_longitude(self, value):
        longitude_sign = {'W': 1, 'E': -1}
        if "W" in value or "E" in value:
            return longitude_sign[value[3]] * float(value[4:])

    def parse_latitude(self, value):
        latitude_sign = {'N': 1, 'S': -1}
        if "N" in value or "S" in value:
            return latitude_sign[value[0]] * float(value[1:3])

    def parse_location(self, column):
        latitude = MaskedColumn(name='Latitude')
        longitude = MaskedColumn(name='Longitude')

        for i, loc in enumerate(column):
            if loc:
                lat_value = self.parse_latitude(loc)
                long_val = self.parse_longitude(loc)
                latitude = latitude.insert(i, lat_value)
                longitude = longitude.insert(i, long_val)
            else:
                latitude = latitude.insert(i, None, mask=True)
                longitude = longitude.insert(i, None, mask=True)
        return latitude, longitude

    def parse_lat_col(self, column, latitude_column):
        for i, loc in enumerate(column):
            if loc:
                latitude_column.mask[i] = False
                latitude_column[i] = self.parse_latitude(loc)
        return latitude_column

    def acquire_data(self, timerange: TimeRange) -> list[str]:
        """
        Function for collecting data with predefined timerange

        :param timerange: TimeRange object with time span for ARs searching
        :type timerange: TimeRange
        :returns: list of urls with txt files.
        :rtype: list of string
        """
        scrapper = Scrapper(self.base_url)
        return scrapper.form_fileslist(timerange)

    def form_data(self, file_urls):
        total_table, section_lines, final_section_lines = [], [], []
        for file_url in file_urls:
            tables = []
            with urlopen(file_url) as response:
                content = response.read().decode('utf-8').split('\n')
                header, section_lines, supplementary_lines = self.extract_lines(content)
                issued_lines = [line for line in header if 'issued' in line.lower() and line.startswith(':')][0]
                _, date_text = issued_lines.strip().split(':')[1:]
                issued_date = datetime.strptime(date_text.strip(), "%Y %b %d %H%M UTC")
                meta_id = OrderedDict()
                for h in header:
                    if h.startswith(("I.", "IA.", "II.")):
                        pos = h.find('.')
                        id = h[:pos]
                        id_text = h[pos + 2:]
                        meta_id[id] = id_text.strip()

                for key, lines in zip(list(meta_id.keys()), section_lines):
                    raw_data = self.proccess_lines(issued_date.strftime("%Y-%m-%d"), key, lines)
                    tables.append(raw_data)
                stacked_table = vstack(tables)

                if 'Location' in stacked_table.columns:
                    col_lat, col_lon = self.parse_location(stacked_table['Location'])
                    del stacked_table['Location']
                    stacked_table.add_column(col_lat)
                    stacked_table.add_column(col_lon)

                if 'Lat' in stacked_table.columns:
                    self.parse_lat_col(stacked_table['Lat'], stacked_table['Latitude'])
                    del stacked_table['Lat']

            total_table.append(stacked_table)
        return Table(vstack(total_table))

    def get_data(self, timerange):
        """
        Function for collecting data with predefined timerange
        :param timerange:
        :type timerange: TimeRange
        :return: data table
        :rtype: Table
        """
        file_urls = self.acquire_data(timerange)
        return self.form_data(file_urls)


class RATANClient(BaseClient):
    """ Client for processing and analyzing data from the RATAN-600 radio telescope.

    RATANClient class is a complex implementation that extends a base client class for handling and processing data
    related to solar observations using RATAN-600. The class provides a wide range of methods to acquire, process,
    and analyze solar data from FITS (Flexible Image Transport System) files obtained from a given URL. Here’s an
    overview of what each method does:

    Key Components:
    1. Initialization Parameters and Configuration:

        - `base_url`: URL template to fetch solar data.
        - regex_pattern: A regex pattern to match relevant data files.
        - convolution_template and quiet_sun_model: Loaded from Excel files; these are templates or models of solar data
        that will be used in the analysis.

    2.Data Acquisition and Filtering:
        - `acquire_data`: Uses a Scrapper class to form a list of data files for a given
        time range, based on URL and regex pattern.
        - `get_scans`: Fetches data from acquired file URLs, processes the FITS
        files to extract relevant information, and returns a table with structured data.

    3. Data Processing and Analysis:

        3.1 Convolution and Efficiency Calculations:

            - convolve_sun: Simulates the convolution of a solar model with Gaussian and rectangular functions, then
              computes and returns the ratio of areas under these convolved functions.
            - antenna_efficiency: Calculates the antenna efficiency across various frequencies using the convolution data.

        3.2 Calibration:

            - calibrate_QSModel: Calibrates the quiet sun model data against observed scan data, adjusting for flux efficiency
              and other parameters

        3.3 Positional and Rotational Transformations:
            - heliocentric_transform: Converts solar latitude and longitude into heliocentric coordinates.
            - pozitional_rotation: Rotates positional data based on a given angle.
            - differential_rotation: Calculates the differential rotation rate of the sun based on latitude.
            - pozitional_angle: Computes the positional angle of the sun based on azimuth, solar declination, and solar P-angle.

        3.4 Active Regions and Peak Analysis:
            - active_regions_search: Identifies active regions on the sun by comparing observed data to known solar
              active region locations using wavelet denoising and peak detection.
            - make_multigauss_fit: Fits multiple Gaussian functions to data points, extracting amplitude, mean, and
              standard deviation for each detected peak.
            - gauss_analysis: Performs Gaussian analysis on identified active regions and associates various metrics
              (amplitude, mean, flux, etc.) with these regions.




    Attributes
    ----------
    base_url : str
        Base URL template for accessing data files.
    regex_pattern : str
        Regular expression pattern to match file names.
    convolution_template : pd.DataFrame
        Template for convolving the quiet Sun model, loaded from the Excel file.
    quiet_sun_model : pd.DataFrame
        Model for the quiet Sun brightness temperature, loaded from the Excel file.
    """

    base_url = 'http://spbf.sao.ru/data/ratan/%Y/%m/%Y%m%d_%H%M%S_sun+0_out.fits'
    regex_pattern = '((\d{6,8})[^0-9].*[^0-9]0_out.fits)'

    convolution_template = pd.read_excel(Path(__file__).absolute().parent.joinpath('quiet_sun_template.xlsx'))
    quiet_sun_model = pd.read_excel(Path(__file__).absolute().parent.joinpath('quiet_sun_model.xlsx'))

    def process_fits_data(self,
                          path_to_file: str,
                          bad_freq: Optional[list[float]] = None,
                          save_path: Optional[str] = None,
                          save_with_original: bool = False,
                          save_raw: bool = False
                          ) -> Tuple[fits.HDUList, fits.HDUList]:
        """
        Process FITS data from a given URL or file_path from the disk, optionally saving the original and processed data to a file,
        and return the processed data as a FITS HDUList object.

        :param path_to_file: The URL of the FITS file to process or path to file at disk.
        :type path_to_file: str
        :param bad_freq: A list of bad frequencies to exclude from the data acquisition.
        :type bad_freq: list of float
        :param save_path: Path to save the processed FITS data. If None, the processed FITS object is returned.
        :type save_path: Optional[str]
        :param save_with_original: Whether to save the original data or not.
        :type save_with_original: bool
        :param save_raw: Whether to save raw FITS data or not.
         :type save_raw: bool
        :returns: The raw and processed FITS data object.
        :rtype: Tuple[fits.HDUList, fits.HDUList]

        :Example:

        .. code-block:: python


            >>> ratan_client = RATANClient()
            >>> raw, processed = ratan_client.process_fits_data('http://spbf.sao.ru/data/ratan/2017/09/20170903_121257_sun+0_out.fits')
            >>> isinstance(processed, fits.HDUList)
            True
        """
        if bad_freq is None:
            bad_freq = [15.0938, 15.2812, 15.4688, 15.6562, 15.8438, 16.0312, 16.2188, 16.4062]
        file_name = str(path_to_file).split('/')[-1]
        file_name_processed = file_name.split('.fits')[0] + '_processed.fits'
        hdul = fits.open(path_to_file)
        hdul.verify('fix')
        data = hdul[0].data

        CDELT1 = hdul[0].header['CDELT1']
        CRPIX = hdul[0].header['CRPIX1']
        SOLAR_R = hdul[0].header['SOLAR_R']
        SOLAR_B = hdul[0].header['SOLAR_B']
        FREQ = hdul[1].data['FREQ']
        OBS_DATE = hdul[0].header['DATE-OBS']
        OBS_TIME = hdul[0].header['TIME-OBS']
        bad_freq = np.isin(FREQ, bad_freq)

        AZIMUTH = hdul[0].header['AZIMUTH']
        SOL_DEC = hdul[0].header['SOL_DEC']
        SOLAR_P = hdul[0].header['SOLAR_P']

        angle = self.pozitional_angle(AZIMUTH, SOL_DEC, SOLAR_P)
        N_shape = data.shape[2]
        x = np.linspace(
            - CRPIX * CDELT1,
            (N_shape - CRPIX) * CDELT1,
            num=N_shape
        )
        flux_eficiency = self.antenna_efficiency(FREQ, SOLAR_R)
        mask, I, V = self.calibrate_QSModel(x, data, SOLAR_R, FREQ, flux_eficiency)
        I, V, FREQ = I[~bad_freq], V[~bad_freq], FREQ[~bad_freq]

        # Pack the processed data into FITS format
        if save_with_original:
            primary_hdu = fits.PrimaryHDU(data)
        else:
            primary_hdu = fits.PrimaryHDU()
        I_hdu = fits.ImageHDU(data=I.astype('float32'), name='I')
        V_hdu = fits.ImageHDU(data=V.astype('float32'), name='V')
        freq_hdu = fits.ImageHDU(data=FREQ.astype('float32'), name='FREQ')
        mask_hdu = fits.ImageHDU(data=mask.astype('int8'), name='mask')

        header = primary_hdu.header
        header['CDELT1'] = CDELT1
        header['CRPIX1'] = CRPIX
        header['SOLAR_R'] = SOLAR_R
        header['SOLAR_B'] = SOLAR_B
        header['DATE-OBS'] = OBS_DATE
        header['TIME-OBS'] = OBS_TIME
        header['AZIMUTH'] = AZIMUTH
        header['SOL_DEC'] = SOL_DEC
        header['SOLAR_P'] = SOLAR_P
        header['ANGLE'] = angle

        hdulist = fits.HDUList([primary_hdu, I_hdu, V_hdu, freq_hdu, mask_hdu, hdul[1]])
        hdulist.verify('fix')

        if save_path:
            hdulist.writeto(Path(save_path) / file_name_processed, overwrite=True)
        if save_raw:
            hdul.writeto(Path(save_path) / file_name, overwrite=True)

        return hdul, hdulist

    def process_fits_with_period(self, timerange: TimeRange,
                                 **kwargs) -> Tuple[List[fits.HDUList], List[fits.HDUList]]:
        """
        Process fits files with a period by urls taken from http://spbf.sao.ru/ and return list of raw
        fits files and processed fits, all kwargs from process_fits_data are available,
         such as option for saving processed fits and raw fits to directory.

        :param timerange: Period of time for which scans will be downloaded and processed
        :param kwargs: keyword arguments valid for process _fits_data
        :return: Tuple[List[fits.HDUList], List[fits.HDUList]]

         :Example:

        .. code-block:: python
        >>> time_range = TimeRange("2017-09-03", "2017-09-04")
        >>> ratan_client = RATANClient()
        >>> _, list_phdul = ratan_client.process_fits_with_period(time_range)
        >>> isinstance(list_phdul, list)
        True

        """
        url_list = self.acquire_data(timerange)
        raw_hdul = []
        processed_hdul = []
        for url in url_list:
            hdul, hdulist = self.process_fits_data(url, **kwargs)
            raw_hdul.append(hdul)
            processed_hdul.append(hdulist)
        return raw_hdul, processed_hdul

    def get_local_sources_info_from_processed(self,
                                              pr_data: Union[str, fits.hdu.hdulist.HDUList],
                                              bad_freq: Optional[list[float]] = None) -> Table:
        """
        Compute local sources info from a processed FITS HDUList and NOAA active regions coordnate info
        :param pr_data:  string with path to file or direct url to source fits or fits file
        :type pr_data: str or fits.hdu.hdulist.HDUList
        :param bad_freq:  list of bd frequencies excluded from computation
        :type bad_freq: list of float
        :return: astropy table with local sources info
        :rtype: Table
        """
        if isinstance(pr_data, str):
            _, processed = self.process_fits_data(pr_data,
                                                  save_path=None,
                                                  save_with_original=False)
        elif isinstance(pr_data, fits.hdu.hdulist.HDUList):
            processed = pr_data
        else:
            NotImplementedError

        if bad_freq is None:
            bad_freq = [15.0938, 15.2812, 15.4688, 15.6562, 15.8438, 16.0312, 16.2188, 16.4062]

        srs_table = self.form_srstable_with_time_shift(processed)

        CDELT1 = processed[0].header['CDELT1']
        CRPIX = processed[0].header['CRPIX1']
        AZIMUTH = processed[0].header['AZIMUTH']
        FREQ = processed[3].data
        bad_freq = np.isin(FREQ, bad_freq)
        I = processed[1].data
        V = processed[2].data
        mask = processed[4].data.astype(bool)
        I, V, FREQ = I[~bad_freq], V[~bad_freq], FREQ[~bad_freq]
        x = np.linspace(
            - CRPIX * CDELT1,
            (V.shape[1] - CRPIX) * CDELT1,
            num=V.shape[1]
        )

        source_info = self.active_regions_search(srs_table, x, V, mask)

        ar_amount = len(source_info)
        source_info.add_column(Column(name='Azimuth', data=[AZIMUTH] * ar_amount, dtype=('i2')), index=1)
        source_info.add_column(
            Column(name='Amplitude', data=[[{} for _ in range(len(FREQ))] for _ in range(ar_amount)], dtype=object))
        source_info.add_column(
            Column(name='Mean', data=[[{} for _ in range(len(FREQ))] for _ in range(ar_amount)], dtype=object))
        source_info.add_column(
            Column(name='Sigma', data=[[{} for _ in range(len(FREQ))] for _ in range(ar_amount)], dtype=object))
        source_info.add_column(
            Column(name='FWHM', data=[[{} for _ in range(len(FREQ))] for _ in range(ar_amount)], dtype=object))
        source_info.add_column(
            Column(name='Flux', data=[[{} for _ in range(len(FREQ))] for _ in range(ar_amount)], dtype=object))
        source_info.add_column(
            Column(name='Total Flux', data=[[{} for _ in range(len(FREQ))] for _ in range(ar_amount)],
                   dtype=object))
        source_info.add_column(
            Column(name='Range', data=[[{} for _ in range(len(FREQ))] for _ in range(ar_amount)], dtype=object))

        ratan_data = list(zip(FREQ, I, V))
        source_info = self.gauss_analysis(x, ratan_data, source_info)

        return source_info

    def get_ar_info_from_processed(self, pr_data: Union[str, fits.hdu.hdulist.HDUList],
                                   bad_freq: Optional[list[float]] = None) -> fits.HDUList:
        """
        Compute ar parameters (total flux, maximum amplitude, width (in arcsec))  info from
        a processed FITS HDUList and NOAA active regions coordnate info
        :param pr_data:  string with path to file or direct url to source fits or fits file
        :type pr_data: str or fits.hdu.hdulist.HDUList
        :param bad_freq:  list of bd frequencies excluded from computation
        :type bad_freq: list of float
        :return: astropy table with local sources info
        :rtype: Table

        >>> raw_fits_data_path = get_project_root() / 'data' / '20170903_121257_sun+0_out.fits'
        >>> ratan_client = RATANClient()
        >>> hdul = ratan_client.get_ar_info_from_processed(str(raw_fits_data_path))
        >>> assert hdul[1].data[0]['Number'] == 2673
        >>> assert hdul[1].data[0]['TotalFlux'].shape == (84,)
        >>> assert hdul[1].data[0]['MaxAmplitude'].shape == (84,)
        >>> assert hdul[1].data[0]['MaxLat'].shape == (84,)
        >>> assert hdul[1].data[0]['MinLat'].shape == (84,)
        """

        assert isinstance(pr_data, (str, pathlib.PosixPath, fits.hdu.hdulist.HDUList)), \
            'Processed data should be link to file, path to fits data of fits data'
        if isinstance(pr_data, (str, pathlib.PosixPath)):
            _, processed = self.process_fits_data(pr_data,
                                                  save_path=None,
                                                  save_with_original=False)
        else:
            processed = pr_data

        if bad_freq is None:
            bad_freq = [15.0938, 15.2812, 15.4688, 15.6562, 15.8438, 16.0312, 16.2188, 16.4062]

        srs_table = self.form_srstable_with_time_shift(processed)

        CDELT1 = processed[0].header['CDELT1']
        CRPIX = processed[0].header['CRPIX1']
        AZIMUTH = processed[0].header['AZIMUTH']
        FREQ = processed[3].data
        bad_freq = np.isin(FREQ, bad_freq)
        I = processed[1].data
        V = processed[2].data
        mask = processed[4].data.astype(bool)
        I, V, FREQ = I[~bad_freq], V[~bad_freq], FREQ[~bad_freq]
        x = np.linspace(
            - CRPIX * CDELT1,
            (V.shape[1] - CRPIX) * CDELT1,
            num=V.shape[1]
        )
        primary_hdu = fits.PrimaryHDU(FREQ)
        primary_hdu.header = processed[0].header
        source_info = self.active_regions_search(srs_table, x, V, mask)

        ar_numbers = np.unique(source_info['Number'])
        total_flux_list = []
        max_amplitude_list = []
        max_x_list = []
        min_x_list = []

        for noaa_ar in ar_numbers:
            region_info = source_info[source_info['Number'] == noaa_ar]
            scan_data = list(zip(FREQ, I, V))
            total_flux = np.zeros_like(FREQ)
            max_amplitude = np.zeros_like(FREQ)
            max_x = np.zeros_like(FREQ)
            min_x = np.zeros_like(FREQ)
            for index, elem in enumerate(scan_data):
                freq, I_data, V_data = elem
                gauss_params, y_min, x_range = self.make_multigauss_fit(x, I_data, region_info)
                gauss_params = gauss_params.reshape(-1, 3)
                gauss_params[:, 0] += y_min
                total_flux[index] = np.sum(np.sqrt(2 * np.pi) * gauss_params[:, 0] * gauss_params[:, 2])
                max_amplitude[index] = np.max(gauss_params[:, 0])
                max_x[index] = np.max(x_range)
                min_x[index] = np.min(x_range)

            total_flux_list.append(total_flux)
            max_amplitude_list.append(max_amplitude)
            max_x_list.append(max_x)
            min_x_list.append(min_x)
        ar_info_part1 = srs_table[['RatanTime', 'Number', 'Mag Type', 'Number of Sunspots', 'Area', 'Z']]
        ar_info_part2 = Table([ar_numbers, total_flux_list, max_amplitude_list, max_x_list, min_x_list],
                              names=('Number', 'TotalFlux', 'MaxAmplitude', 'MaxLat', 'MinLat'))
        ar_info = join(ar_info_part1, ar_info_part2, keys='Number')
        return fits.HDUList([primary_hdu, fits.BinTableHDU(ar_info)])
    
    def find_ar_intervals_using_peaks(self, x: np.ndarray, 
                                    y: np.ndarray, 
                                    ar_table: Table) -> Table:
        """
        Identify intervals around active region centers using peak detection

        :param x: Array of x-coordinates, typically representing time or position
        :type x: np.ndarray
        :param y: Array of y-coordinates, representing the signal intensity
        :type y: np.ndarray
        :param ar_table: Table with active region info, including 'Number' and 'Latitude'
        :type ar_table: Table
        :return: Table with active region numbers and calculated intervals
        :rtype: Table
        """
        peaks_indices, _ = find_peaks(y)
        minima_indices, _ = find_peaks(-y)

        ar_intervals = []
        for number, center_x in ar_table[['Number', 'Latitude']]:
            peak_distances = np.abs(x[peaks_indices] - center_x)
            center_index = peaks_indices[np.argmin(peak_distances)]

            left_minima = minima_indices[minima_indices < center_index]
            right_minima = minima_indices[minima_indices > center_index]

            if len(left_minima) > 0:
                left_index = left_minima[-1]
            else:
                left_index = 0 

            if len(right_minima) > 0:
                right_index = right_minima[0]
            else:
                right_index = len(y) - 1 

            ar_interval = (x[left_index], x[right_index])
            ar_intervals.append(ar_interval)

        ar_table_with_intervals = ar_table.copy()
        ar_table_with_intervals.add_column(Column(data=ar_intervals, name='Interval'))
        return ar_table_with_intervals
    
    def compute_fluxes(self, x: np.ndarray, 
                       y: np.ndarray, 
                       ar_table: Table, 
                       mode: str = 'I') -> Table:
        """
        Calculate fluxes for active regions over specified intervals.

        :param x: Array of x-coordinates, typically representing time or position.
        :type x: np.ndarray
        :param y: Array of y-coordinates, representing the signal intensity. Supports 2D arrays.
        :type y: np.ndarray
        :param ar_table: Table containing active region info, including 'Interval'.
        :type ar_table: Table
        :param mode: Mode of flux calculation, defaults to 'I'.
        :type mode: str
        :return: Table with active region info and calculated fluxes.
        :rtype: Table
        """
        y = np.atleast_2d(y)
        cumulative_integral = cumulative_trapezoid(y, x, axis=1, initial=0)

        cum_integral_interp = interp1d(
            x, cumulative_integral, kind='linear', axis=1,
            fill_value="extrapolate", bounds_error=False
        )

        intervals = np.array(ar_table['Interval'])
        left_x = intervals[:, 0].astype(np.float64)
        right_x = intervals[:, 1].astype(np.float64)

        left_x, right_x = np.minimum(left_x, right_x), np.maximum(left_x, right_x)

        cum_int_left = cum_integral_interp(left_x)
        cum_int_right = cum_integral_interp(right_x)

        fluxes = (cum_int_right - cum_int_left)

        col_name = 'Flux_' + mode
        flux_column = Column(data=fluxes.T, name=col_name)
        ar_table_with_fluxes = ar_table.copy()
        ar_table_with_fluxes.add_column(flux_column)
        return ar_table_with_fluxes
    
    def get_ar_info(self, 
                    pr_data: Union[str, fits.hdu.hdulist.HDUList], 
                    **kwargs) -> Table:
        """
        Retrieve Extract active region information from processed FITS data.
        
        :param pr_data: Path to FITS file, URL, or FITS HDUList with processed data.
        :type pr_data: Union[str, fits.hdu.hdulist.HDUList]
        :param kwargs: Additional keyword arguments for processing options.
        :return: FITS HDUList containing frequency data and active region information.
        :rtype: Table
        """
        assert isinstance(pr_data, (str, pathlib.PosixPath, fits.hdu.hdulist.HDUList)), \
            'Processed data should be link to file, path to fits data of fits data'
        if isinstance(pr_data, (str, pathlib.PosixPath)):
            _, processed = self.process_fits_data(pr_data,
                                                  save_path=None,
                                                  save_with_original=False)
        else:
            processed = pr_data

        if bad_freq is None:
            bad_freq = [15.0938, 15.2812, 15.4688, 15.6562, 15.8438, 16.0312, 16.2188, 16.4062]

        srs_table = self.form_srstable_with_time_shift(processed)

        CDELT1 = processed[0].header['CDELT1']
        CRPIX = processed[0].header['CRPIX1']
        FREQ = processed[3].data
        bad_freq = np.isin(FREQ, bad_freq)
        I = processed[1].data
        V = processed[2].data
        mask = processed[4].data.astype(bool)
        I, V, FREQ = I[~bad_freq], V[~bad_freq], FREQ[~bad_freq]
        x = np.linspace(
            - CRPIX * CDELT1,
            (V.shape[1] - CRPIX) * CDELT1,
            num=V.shape[1]
        )

        primary_hdu = fits.PrimaryHDU(FREQ)
        primary_hdu.header = processed[0].header
        
        ar_intervals = self.find_ar_intervals_using_peaks(x[mask], I[0][mask], srs_table)
        ar_info = self.compute_fluxes(x, I, ar_intervals, mode='I')
        ar_info = self.compute_fluxes(x, V, ar_intervals, mode='V')
        return fits.HDUList([primary_hdu, fits.BinTableHDU(ar_info)])

    def condition(self, year, month, data_match):
        """
        Formats the data string based on the year and month. Used in Scrapper class
        :param year: The year of the data.
        :type year: str
        :param month:
        :type month: str
        :param data_match: data text from http content
        :type data_match: str
        :return: string with selected data
        """
        if int(year) < 2010 or (int(year) == 2010 and int(month) < 5):
            return f'{year[:2]}{data_match[:-4]}-{data_match[-4:-2]}-{data_match[-2:]}'
        else:
            return f'{data_match[:-4]}-{data_match[-4:-2]}-{data_match[-2:]}'

    def filter(self, url_list):
        pass

    @lru_cache(maxsize=None)
    def convolve_sun(self, sigma_horiz: float, sigma_vert: float, R: float):
        """    Performs convolution of the Sun model with Gaussian and rectangular functions.

        :param sigma_horiz:
        :type sigma_horiz: float
        :param sigma_vert:
        :type sigma_vert: float
        :param R:
        :type R: float
        :return:
        """
        size = 1000
        x = np.linspace(-size // 2, size // 2, size)
        y = np.linspace(-size // 2, size // 2, size)
        gaussian = gauss2d(x, y, 1, 1, 0, 0, sigma_horiz, sigma_vert)
        rectangle = create_rectangle(size, 6 * sigma_horiz, 4 * R)
        sun_model = create_sun_model(size, R)
        # Perform convolutions
        convolved_gaussian = fftconvolve(sun_model, gaussian, mode='same', axes=1) / np.sum(gaussian)
        convolved_rectangle = fftconvolve(sun_model, rectangle, mode='same', axes=1) / np.sum(rectangle)

        convolved_gaussian = convolved_gaussian / np.max(convolved_gaussian)
        convolved_rectangle = convolved_rectangle / np.max(convolved_rectangle)

        conv_g = np.sum(convolved_gaussian, axis=0)
        conv_r = np.sum(convolved_rectangle, axis=0)
        # Calculate areas under the curve
        area_gaussian = calculate_area(conv_g)
        area_rectangle = calculate_area(conv_r)
        # Division of areas
        area_ratio = area_gaussian / area_rectangle
        return area_ratio

    def antenna_efficiency(self, freq: List[float], R: float) -> List[float]:
        """
        Calculates the antenna efficiency for each frequency.

        :param freq:
        :param R:
        :return:
        """
        areas = []
        for f in freq:
            lambda_value = (3 * 10 ** 8) / (f * 10 ** 9) * 1000
            FWHM_h = 0.85 * lambda_value
            FWHM_v = 0.75 * lambda_value * 60

            sigma_h = round(bwhm_to_sigma(FWHM_h), 9)
            sigma_v = round(bwhm_to_sigma(FWHM_v), 9)

            area_ratio = self.convolve_sun(sigma_h, sigma_v, R)
            areas.append(area_ratio)
        return areas

    def calibrate_QSModel(self, x: np.ndarray, scan_data: np.ndarray,
                          solar_r: float, frequency: np.ndarray,
                          flux_eficiency: np.ndarray) -> Tuple[np.ndarray, np.ndarray, np.ndarray]:
        """
        Calibrates raw scan data with the quiet Sun model.
        :param x: Array of x coordinates
        :type: np.ndarray
        :param scan_data: Scan data array.
        :type: np.ndarray
        :param solar_r:  Solar radius in arcseconds
        :type: float
        :param frequency: Array of frequencies.
        :type: np.ndarray
        :param flux_eficiency: Array of flux efficiencies for each frequency.
        :type: np.ndarray
        :return: calibrated scan: mask, Stocks component I, Stocks component V
        """
        K_b = 1.38 * 10 ** (-23)  # Boltzmann constant
        c = 3 * 10 ** 8  # Speed of light in vacuum
        freq_size = len(frequency)
        model_freq = self.convolution_template.columns[1:].values.astype(float)

        template_freq = self.quiet_sun_model['freq']
        template_val = self.quiet_sun_model['T_brightness']
        real_brightness = interp1d(template_freq, template_val, bounds_error=False, fill_value="extrapolate")
        full_flux = np.column_stack([
            frequency,
            2 * 10 ** 22 * K_b * real_brightness(frequency) * SunSolidAngle(solar_r) * (c / (frequency * 10 ** 9)) ** (
                -2)
        ])

        R = scan_data[:, 0, :] + scan_data[:, 1, :]
        L = scan_data[:, 0, :] - scan_data[:, 1, :]

        columns = self.convolution_template.columns.values[1:].astype(float)
        mask = (x >= -1.0 * solar_r) & (x <= 1.0 * solar_r)

        calibrated_R = np.zeros((freq_size, R.shape[1]), dtype=float)
        calibrated_L = np.zeros((freq_size, L.shape[1]), dtype=float)
        theoretical_new = np.zeros((freq_size, L.shape[1]), dtype=float)

        for freq_num in range(freq_size):
            real_R, real_L = R[freq_num], L[freq_num]
            freq_diff = np.abs(columns - frequency[freq_num])
            freq_template = np.argmin(freq_diff) + 1

            template_values = self.convolution_template.iloc[:, freq_template].values.copy()
            x_values = self.convolution_template.iloc[:, 0].values.copy()

            convolution_template_arcsec = flip_and_concat(template_values)
            x_arcsec = flip_and_concat(x_values, flip_values=True)

            coeff = full_flux[freq_num, 1] * flux_eficiency[freq_num] / trapezoid(convolution_template_arcsec, x_arcsec)
            theoretical_data = np.interp(x, x_arcsec, coeff * convolution_template_arcsec)
            theoretical_new[freq_num] = theoretical_data

            res_R = minimize(error, np.array([1]), args=(real_R[mask], theoretical_data[mask]))
            res_L = minimize(error, np.array([1]), args=(real_L[mask], theoretical_data[mask]))

            calibrated_R[freq_num, :] = real_R * res_R.x
            calibrated_L[freq_num, :] = real_L * res_L.x
        return mask, (calibrated_R + calibrated_L) / 2, (calibrated_R - calibrated_L) / 2

    def heliocentric_transform(self, Lat: np.ndarray, Long: np.ndarray,
                               SOLAR_R: float, SOLAR_B: float) -> Tuple[np.ndarray, np.ndarray]:
        """
        Transforms solar coordinates to a heliocentric system.
        :param Lat:  Latitude array in degrees
        :type: np.ndarray
        :param Long: Longitude array in degrees
        :type: np.ndarray
        :param SOLAR_R:  Solar radius.
        :type: float
        :param SOLAR_B: Solar B angle
        :type: float
        :return: Heliocentric x and y coordinates.
        :rtype: Tuple[np.ndarray, np.ndarray]

        """
        return (
            SOLAR_R * np.cos(Lat * np.pi / 180) * np.sin(Long * np.pi / 180),
            SOLAR_R * (np.sin(Lat * np.pi / 180) * np.cos(SOLAR_B * np.pi / 180)
                       - np.cos(Lat * np.pi / 180) * np.cos(Long * np.pi / 180) * np.sin(SOLAR_B * np.pi / 180))
        )

    def pozitional_rotation(self, Lat: np.ndarray, Long: np.ndarray,
                            angle: float) -> Tuple[np.ndarray, np.ndarray]:
        """
        Rotates positional data by a given angle.

        :param Lat:  Latitude array in degrees
        :type: np.ndarray
        :param Long: Longitude array in degrees
        :type: np.ndarray
        :param angle:    Rotation angle in degrees.
        :type: float
        :return: Rotated X and Y coordinates
        :rtype: Tuple[np.ndarray, np.ndarray]
        """
        return (
            Lat * np.cos(angle * np.pi / 180) - Long * np.sin(angle * np.pi / 180),
            Lat * np.sin(angle * np.pi / 180) + Long * np.cos(angle * np.pi / 180)
        )

    def differential_rotation(self, Lat: np.ndarray) -> np.ndarray:
        """
        Calculates the differential rotation of the Sun based on latitude.
        :param Lat:  Latitude array in degrees
        :type: np.ndarray
        :return:
        """
        A = 14.713
        B = -2.396
        C = -1.787
        return A + B * np.sin(Lat * np.pi / 180) ** 2 + C * np.sin(Lat * np.pi / 180) ** 4

    def pozitional_angle(self, AZIMUTH: float, SOL_DEC: float, SOLAR_P: float) -> float:
        """
         Calculates the positional angle of the sun based on azimuth, solar declination, and solar P-angle.
        :param AZIMUTH: Azimuth angle.
        :type: float
        :param SOL_DEC:  Solar declination angle.
        :type: float
        :param SOLAR_P: Solar P-angle
        :type: float
        :return: Positional angle of the sun
        :rtype: float
        """
        q = -np.arcsin(np.tan(AZIMUTH * np.pi / 180) * np.tan(SOL_DEC * np.pi / 180)) * 180 / np.pi
        p = SOLAR_P + 360.0 if np.abs(SOLAR_P) > 30 else SOLAR_P
        return (p + q)

    def form_srstable_with_time_shift(self, processed_file: fits.HDUList) -> Table:
        """ Create table with AR info with SRSClient
        with correct NOAA coordinates for RATAN time difference

        :param processed_file:  calibrate RATAN fits
        :type: fits.HDUList
        :return: modified NOAA table with ARs
        :rtype: Table
        """

        SOLAR_R = processed_file[0].header['SOLAR_R'] * 1.01175
        SOLAR_B = processed_file[0].header['SOLAR_B']
        OBS_DATE = processed_file[0].header['DATE-OBS']
        OBS_TIME = processed_file[0].header['TIME-OBS']

        AZIMUTH = processed_file[0].header['AZIMUTH']
        SOL_DEC = processed_file[0].header['SOL_DEC']
        SOLAR_P = processed_file[0].header['SOLAR_P']
        angle = self.pozitional_angle(AZIMUTH, SOL_DEC, SOLAR_P)
        ratan_datetime = datetime.strptime(OBS_DATE + ' ' + OBS_TIME, '%Y/%m/%d %H:%M:%S.%f')
        ratan_datetime_str = ratan_datetime.strftime('%Y/%m/%d %H:%M')
        noaa_datetime = datetime.strptime(OBS_DATE, '%Y/%m/%d')
        diff_hours = int((ratan_datetime - noaa_datetime).total_seconds() / 3600)

        srs = SRSClient()
        srs_table = srs.get_data(TimeRange(OBS_DATE, OBS_DATE))
        srs_table = srs_table[srs_table['ID'] == 'I']
        len_tbl = len(srs_table)
        srs_table.add_column(Column(name='RatanTime', data=[ratan_datetime_str] * len_tbl))
        srs_table['Longitude'] = (
                srs_table['Longitude'] + self.differential_rotation(srs_table['Latitude']) * diff_hours / 24
        ).astype(int)

        srs_table['Latitude'], srs_table['Longitude'] = self.heliocentric_transform(
            srs_table['Latitude'],
            srs_table['Longitude'],
            SOLAR_R,
            SOLAR_B
        )

        srs_table['Latitude'], srs_table['Longitude'] = self.pozitional_rotation(
            srs_table['Latitude'],
            srs_table['Longitude'],
            angle
        )

        return srs_table

    def active_regions_search(self, srs: Table, x: np.ndarray, V: np.ndarray, mask: np.ndarray):
        """
         Searches for active regions in the solar spectrum data using wavelet denoising and peak detection
         and matching with NOAA active location.
        :param srs: table from SRSCLient with informations about the NOAA active regions.
        :param x: x coordinates
        :param V: Stocks vector V
        :param mask: masked interval
        :return: Table with active regions: V value and coordinate for center of AR
        :rtype: Table
        """
        x = x[mask]
        V = np.sum(V, axis=0)[mask]

        wavelet = 'sym6'  # Daubechies wavelet
        level = 4  # Level of decomposition
        denoised_data = wavelet_denoise(V, wavelet, level)
        height_threshold = lambda x: np.abs(np.median(x) + 0.1 * np.std(x))
        # Finding peaks in the denoised data
        peaks, _ = find_peaks(denoised_data, height=height_threshold(denoised_data))
        valleys, _ = find_peaks(-denoised_data, height=height_threshold(-denoised_data))
        extremums = np.concatenate((peaks, valleys))

        theoretical_latitudes = np.array(srs['Latitude'])
        experimental_latitudes = x[extremums]
        abs_diff = np.abs(experimental_latitudes[:, np.newaxis] - theoretical_latitudes)
        min_index = np.argmin(abs_diff, axis=1)
        closest_data = srs[min_index]
        closest_data['Latitude'] = experimental_latitudes

        original_indices = np.where(mask)[0]
        extremums_original = original_indices[extremums]
        closest_data.add_column(Column(name='Data Index', data=extremums_original, dtype=('i4')), index=3)
        closest_data.add_column(Column(name='Masked Index', data=extremums, dtype=('i4')), index=4)
        return closest_data

    def make_multigauss_fit(self, x: np.ndarray, y: np.ndarray,
                            peak_info: Table) -> Tuple[float, float, float]:
        """
        Fits multiple Gaussian functions to detected peaks.
        :param x: x values of the data.
        :type: np.ndarray
        :param y: y values of the data.
        :type: np.ndarray
        :param peak_info:  astropy Table with information about peaks.
        :return: width, value and coordinate for center of AR.
        :rtype: Tuple[float, float, float]
        """
        min_lat = np.min(peak_info['Latitude'])
        max_lat = np.max(peak_info['Latitude'])
        indexes = peak_info['Data Index']
        mask = (x >= min_lat - 50) & (x <= max_lat + 50)
        x_masked, y_masked = x[mask], y[mask]
        y_min = np.min(y_masked)
        ar_info = [[y[index] - y_min, x[index]] for index in indexes]
        widths = np.repeat(1, len(peak_info))
        initial_guesses = np.ravel(np.column_stack((ar_info, widths)))
        params, _ = leastsq(gaussian_mixture, initial_guesses, args=(x_masked, y_masked - y_min))
        return np.array(params), y_min, x_masked

    def gauss_analysis(self, x, scan_data, ar_info):
        """
        Analyzes active regions by fitting Gaussian functions and calculating associated metrics.
        :param x: x coordinates of the data (arcsec).
        :param scan_data:  scan with radio data
        :type: np.ndarray
        :param ar_info: Table with information about active regions.
        :type: Table
        :return: Table with ARs parameters from Radio data at each frequency
        :rtype: Table
        """
        ar_number = np.unique(ar_info['Number'])
        for noaa_ar in ar_number:
            region_info = ar_info[ar_info['Number'] == noaa_ar]
            indices = np.where(ar_info['Number'] == noaa_ar)[0]
            for index, elem in enumerate(scan_data):
                freq, I_data, V_data = elem
                gauss_params, y_min, x_range = self.make_multigauss_fit(x, I_data, region_info)
                gauss_params = gauss_params.reshape(-1, 3)
                gauss_params[:, 0] += y_min
                total_flux = np.sum(np.sqrt(2 * np.pi) * gauss_params[:, 0] * gauss_params[:, 2])
                for local_index, gaussian in zip(indices, gauss_params):
                    amplitude, mean, stddev = gaussian
                    ar_info['Amplitude'][local_index][index] = {'freq': freq, 'amplitude': amplitude}
                    ar_info['Mean'][local_index][index] = {'freq': freq, 'mean': mean}
                    ar_info['Sigma'][local_index][index] = {'freq': freq, 'sigma': stddev}
                    ar_info['FWHM'][local_index][index] = {'freq': freq, 'fwhm': 2 * np.sqrt(2 * np.log(2)) * stddev}
                    ar_info['Range'][local_index][index] = {'freq': freq, 'x_range': (np.min(x_range), np.max(x_range))}
                    ar_info['Flux'][local_index][index] = {'freq': freq,
                                                           'flux': np.sqrt(2 * np.pi) * amplitude * stddev}
                    ar_info['Total Flux'][local_index][index] = {'freq': freq, 'flux': total_flux}
        return ar_info

    def acquire_data(self, timerange: TimeRange) -> List[str]:
        """
        Get all available urls for predefined TimeRange
        :param timerange: timerange for scan searching
        :type: TimeRange
        :return: url list:
        :rtype:
        """
        scrapper = Scrapper(self.base_url, regex_pattern=self.regex_pattern,
                            condition=self.condition, filter=self.filter)
        return scrapper.form_fileslist(timerange)

    def get_scans(self, timerange):
        file_urls = self.acquire_data(timerange)
        column_types = {
            'Date': np.dtype('U10'),
            'Time': np.dtype('U8'),
            'Azimuth': np.dtype('i2'),
            'SOLAR_R': np.dtype('float64'),
            'N_shape': np.dtype('i4'),
            'CRPIX': np.dtype('float64'),
            'CDELT1': np.dtype('float64'),
            'Pozitional Angle': np.dtype('float64'),
            'SOLAR_B': np.dtype('float64'),
            'Frequency': np.dtype('O'),
            'Flux Eficiency ': np.dtype('O'),
            'I': np.dtype('O'),
            'V': np.dtype('O'),
        }

        table = Table(names=tuple(column_types.keys()), dtype=tuple(column_types.values()))
        for file_url in file_urls:
            hdul = fits.open(file_url)
            data = hdul[0].data

            CDELT1 = hdul[0].header['CDELT1']
            CRPIX = hdul[0].header['CRPIX1']
            SOLAR_R = hdul[0].header['SOLAR_R'] * 1.01175
            SOLAR_B = hdul[0].header['SOLAR_B']
            FREQ = hdul[1].data['FREQ']
            OBS_DATE = hdul[0].header['DATE-OBS']
            OBS_TIME = hdul[0].header['TIME-OBS']
            bad_freq = np.isin(FREQ, [15.0938, 15.2812, 15.4688, 15.6562, 15.8438, 16.0312, 16.2188, 16.4062])

            AZIMUTH = hdul[0].header['AZIMUTH']
            SOL_DEC = hdul[0].header['SOL_DEC']
            SOLAR_P = hdul[0].header['SOLAR_P']
            angle = self.pozitional_angle(AZIMUTH, SOL_DEC, SOLAR_P)
            N_shape = data.shape[2]
            x = np.linspace(
                - CRPIX * CDELT1,
                (N_shape - CRPIX) * CDELT1,
                num=N_shape
            )
            flux_eficiency = self.antenna_efficiency(FREQ, SOLAR_R)
            mask, I, V = self.calibrate_QSModel(x, data, SOLAR_R, FREQ, flux_eficiency)
            I, V, FREQ = I[~bad_freq], V[~bad_freq], FREQ[~bad_freq]
            table.add_row([
                OBS_DATE.replace('/', '-'),
                OBS_TIME,
                AZIMUTH,
                SOLAR_R,
                N_shape,
                CRPIX,
                CDELT1,
                angle,
                SOLAR_B,
                FREQ,
                dict(zip(FREQ, flux_eficiency)),
                dict(zip(FREQ, I)),
                dict(zip(FREQ, V))
            ])
        return table

    def form_data(self, file_urls):
        total_table = []
        for file_url in file_urls:
            hdul = fits.open(file_url)
            data = hdul[0].data

            CDELT1 = hdul[0].header['CDELT1']
            CRPIX = hdul[0].header['CRPIX1']
            SOLAR_R = hdul[0].header['SOLAR_R'] * 1.01175
            SOLAR_B = hdul[0].header['SOLAR_B']
            FREQ = hdul[1].data['FREQ']
            OBS_DATE = hdul[0].header['DATE-OBS']
            OBS_TIME = hdul[0].header['TIME-OBS']
            bad_freq = np.isin(FREQ, [15.0938, 15.2812, 15.4688, 15.6562, 15.8438, 16.0312, 16.2188, 16.4062])

            ratan_datetime = datetime.strptime(OBS_DATE + ' ' + OBS_TIME, '%Y/%m/%d %H:%M:%S.%f')
            noaa_datetime = datetime.strptime(OBS_DATE, '%Y/%m/%d')
            diff_hours = int((ratan_datetime - noaa_datetime).total_seconds() / 3600)

            AZIMUTH = hdul[0].header['AZIMUTH']
            SOL_DEC = hdul[0].header['SOL_DEC']
            SOLAR_P = hdul[0].header['SOLAR_P']
            angle = self.pozitional_angle(AZIMUTH, SOL_DEC, SOLAR_P)

            x = np.linspace(
                - CRPIX * CDELT1,
                (data.shape[2] - CRPIX) * CDELT1,
                num=data.shape[2]
            )

            flux_eficiency = self.antenna_efficiency(FREQ, SOLAR_R)
            mask, I, V = self.calibrate_QSModel(x, data, SOLAR_R, FREQ, flux_eficiency)
            I, V, FREQ = I[~bad_freq], V[~bad_freq], FREQ[~bad_freq]

            srs = SRSClient()
            srs_table = srs.get_data(TimeRange(OBS_DATE, OBS_DATE))
            srs_table = srs_table[srs_table['ID'] == 'I']
            srs_table['Longitude'] = (
                    srs_table['Longitude'] + self.differential_rotation(srs_table['Latitude']) * diff_hours / 24
            ).astype(int)

            srs_table['Latitude'], srs_table['Longitude'] = self.heliocentric_transform(
                srs_table['Latitude'],
                srs_table['Longitude'],
                SOLAR_R,
                SOLAR_B
            )

            srs_table['Latitude'], srs_table['Longitude'] = self.pozitional_rotation(
                srs_table['Latitude'],
                srs_table['Longitude'],
                angle
            )

            ar_info = self.active_regions_search(srs_table, x, V, mask)
            ar_amount = len(ar_info)
            ar_info.add_column(Column(name='Azimuth', data=[AZIMUTH] * ar_amount, dtype=('i2')), index=1)
            ar_info.add_column(
                Column(name='Amplitude', data=[[{} for _ in range(len(FREQ))] for _ in range(ar_amount)], dtype=object))
            ar_info.add_column(
                Column(name='Mean', data=[[{} for _ in range(len(FREQ))] for _ in range(ar_amount)], dtype=object))
            ar_info.add_column(
                Column(name='Sigma', data=[[{} for _ in range(len(FREQ))] for _ in range(ar_amount)], dtype=object))
            ar_info.add_column(
                Column(name='FWHM', data=[[{} for _ in range(len(FREQ))] for _ in range(ar_amount)], dtype=object))
            ar_info.add_column(
                Column(name='Flux', data=[[{} for _ in range(len(FREQ))] for _ in range(ar_amount)], dtype=object))
            ar_info.add_column(
                Column(name='Total Flux', data=[[{} for _ in range(len(FREQ))] for _ in range(ar_amount)],
                       dtype=object))
            ar_info.add_column(
                Column(name='Range', data=[[{} for _ in range(len(FREQ))] for _ in range(ar_amount)], dtype=object))

            ratan_data = list(zip(FREQ, I, V))
            gauss_analysis_info = self.gauss_analysis(x, ratan_data, ar_info)
            total_table.append(gauss_analysis_info)
        return Table(vstack(total_table))

    def get_data(self, timerange):
        file_urls = self.acquire_data(timerange)
        return self.form_data(file_urls)
